#
# Copyright (C) 2021 FreeIPA Contributors see COPYING for license
#

from util import capture_results, m_api, CAInstance, KRAInstance
from base import BaseTest
from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.certs import IPACertMatchCheck
from ipahealthcheck.ipa.certs import IPADogtagCertsMatchCheck
from unittest.mock import Mock, patch

from ipalib import errors
from ipapython.dn import DN
from ipapython.ipaldap import LDAPClient, LDAPEntry


class IPACertificate:
    def __init__(self, serial_number=1):
        self.serial_number = serial_number

    def __eq__(self, other):
        return self.serial_number == other.serial_number

    def __hash__(self):
        return hash(self.serial_number)


class mock_ldap:
    SCOPE_BASE = 1
    SCOPE_ONELEVEL = 2
    SCOPE_SUBTREE = 4

    def __init__(self, entries):
        """Initialize the results that we will return from get_entry"""
        self.results = {entry.dn: entry for entry in entries}

    def get_entry(self, dn, attrs_list=None, time_limit=None,
                  size_limit=None, get_effective_rights=False):
        if self.results is None:
            raise errors.NotFound(reason='test')
        return self.results[dn]

    def get_entries(self, base_dn, scope=SCOPE_SUBTREE, filter=None,
                    attrs_list=None, get_effective_rights=False, **kwargs):
        if self.results is None:
            raise errors.NotFound(reason='None')
        if filter:
            (attr, value) = filter.split('=', maxsplit=1)
            for result in self.results.values():
                if result.get(attr)[0] == value:
                    return [result]
            raise errors.NotFound(reason='Not found %s' % filter)

        return self.results


class mock_ldap_conn:
    def set_option(self, option, invalue):
        pass

    def search_s(self, base, scope, filterstr=None,
                 attrlist=None, attrsonly=0):
        return tuple()


class mock_CertDB:
    def __init__(self, trust):
        """A dict of nickname + NSSdb trust flags"""
        self.trust = trust
        self.secdir = '/foo/bar/testdir'

    def get_cert_from_db(self, nickname):
        if nickname not in self.trust.keys():
            raise errors.NotFound(reason='test')
        return IPACertificate()

    def run_certutil(self, args, capture_output):
        class RunResult:
            def __init__(self, output):
                self.raw_output = output

        return RunResult(b'test output')


class TestIPACertMatch(BaseTest):
    patches = {
        'ldap.initialize':
        Mock(return_value=mock_ldap_conn())
    }

    trust = {
        ('%s IPA CA' % m_api.env.realm): 'u,u,u'
    }

    @patch('ipalib.x509.load_certificate_list_from_file')
    @patch('ipaserver.install.certs.CertDB')
    def test_certs_match_ok(self, mock_certdb, mock_load_cert):
        """ Ensure match check is ok"""
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        cacertentry = LDAPEntry(fake_conn,
                                DN('cn=%s IPA CA' % m_api.env.realm,
                                   'cn=certificates,cn=ipa,cn=etc',
                                    m_api.env.basedn),
                                CACertificate=[IPACertificate()])

        mock_certdb.return_value = mock_CertDB(self.trust)
        mock_load_cert.return_value = [IPACertificate()]

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPACertMatchCheck(registry)
        f.conn = mock_ldap([cacertentry])
        self.results = capture_results(f)

        assert len(self.results) == 3
        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPACertMatchCheck'

    @patch('ipalib.x509.load_certificate_list_from_file')
    @patch('ipaserver.install.certs.CertDB')
    def test_etc_cacert_mismatch(self, mock_certdb, mock_load_cert):
        """ Test mismatch with /etc/ipa/ca.crt """
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        cacertentry = LDAPEntry(fake_conn,
                                DN('cn=%s IPA CA' % m_api.env.realm,
                                   'cn=certificates,cn=ipa,cn=etc',
                                    m_api.env.basedn),
                                CACertificate=[IPACertificate()])

        mock_certdb.return_value = mock_CertDB(self.trust)
        mock_load_cert.return_value = [IPACertificate(serial_number=2)]

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPACertMatchCheck(registry)
        f.conn = mock_ldap([cacertentry])
        self.results = capture_results(f)

        assert len(self.results) == 3
        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPACertMatchCheck'

    @patch('ipaserver.install.cainstance.CAInstance')
    def test_cacert_caless(self, mock_cainstance):
        """Nothing to check if the master is CALess"""

        mock_cainstance.return_value = CAInstance(False)

        framework = object()
        registry.initialize(framework, config)
        f = IPACertMatchCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 0


default_subject_base = [{
    'result':
        {
            'ipacertificatesubjectbase': [f'O={m_api.env.realm}'],
        },
}]

custom_subject_base = [{
    'result':
        {
            'ipacertificatesubjectbase': ['OU=Eng,O=ACME'],
        },
}]


class TestIPADogtagCertMatch(BaseTest):
    patches = {
        'ipaserver.install.krainstance.KRAInstance':
        Mock(return_value=KRAInstance()),
    }
    trust = {
        'ocspSigningCert cert-pki-ca': 'u,u,u',
        'caSigningCert cert-pki-ca': 'u,u,u',
        'subsystemCert cert-pki-ca': 'u,u,u',
        'auditSigningCert cert-pki-ca': 'u,u,Pu',
        'Server-Cert cert-pki-ca': 'u,u,u',
        'transportCert cert-pki-kra': 'u,u,u',
        'storageCert cert-pki-kra': 'u,u,u',
        'auditSigningCert cert-pki-kra': 'u,u,Pu',
    }

    def get_dogtag_subjects(self, hostname, base):
        subject_base = base[0]['result']['ipacertificatesubjectbase'][0]
        return (
            f'CN=OCSP Subsystem,{subject_base}',
            f'CN=CA Subsystem,{subject_base}',
            f'CN=CA Audit,{subject_base}',
            f'CN=%s,{subject_base}',
            f'CN=KRA Transport Certificate,{subject_base}',
            f'CN=KRA Storage Certificate,{subject_base}',
            f'CN=KRA Audit,{subject_base}',
            f'CN={hostname},{subject_base}',
        )

    @patch('ipaserver.install.certs.CertDB')
    def test_certs_match_ok(self, mock_certdb):
        """ Ensure match check is ok"""
        m_api.Command.config_show.side_effect = default_subject_base
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        pkidbentry = LDAPEntry(fake_conn,
                               DN('uid=pkidbuser,ou=people,o=ipaca'),
                               userCertificate=[IPACertificate()],
                               subjectName=['test'])
        casignentry = LDAPEntry(fake_conn,
                                DN('cn=%s IPA CA' % m_api.env.realm,
                                   'cn=certificates,cn=ipa,cn=etc',
                                    m_api.env.basedn),
                                CACertificate=[IPACertificate()],
                                userCertificate=[IPACertificate()],
                                subjectName=['test'])
        ldap_entries = [pkidbentry, casignentry]

        dogtag_entries_subjects = self.get_dogtag_subjects(
            m_api.env.host, default_subject_base
        )

        for i, subject in enumerate(dogtag_entries_subjects):
            entry = LDAPEntry(fake_conn,
                              DN('cn=%i,ou=certificateRepository' % i,
                                 'ou=ca,o=ipaca'),
                              userCertificate=[IPACertificate()],
                              subjectName=[subject])
            ldap_entries.append(entry)

        mock_certdb.return_value = mock_CertDB(self.trust)

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPADogtagCertsMatchCheck(registry)
        f.conn = mock_ldap(ldap_entries)
        self.results = capture_results(f)

        assert len(self.results) == 3
        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPADogtagCertsMatchCheck'

    @patch('ipaserver.install.certs.CertDB')
    def test_certs_mismatch(self, mock_certdb):
        """ Ensure mismatches are detected"""
        m_api.Command.config_show.side_effect = default_subject_base
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        pkidbentry = LDAPEntry(fake_conn,
                               DN('uid=pkidbuser,ou=people,o=ipaca'),
                               userCertificate=[IPACertificate(
                                   serial_number=2
                               )],
                               subjectName=['test'])
        casignentry = LDAPEntry(fake_conn,
                                DN('cn=%s IPA CA' % m_api.env.realm,
                                   'cn=certificates,cn=ipa,cn=etc',
                                    m_api.env.basedn),
                                CACertificate=[IPACertificate()],
                                userCertificate=[IPACertificate()],
                                subjectName=['test'])
        ldap_entries = [pkidbentry, casignentry]

        dogtag_entries_subjects = self.get_dogtag_subjects(
            m_api.env.host, default_subject_base
        )

        for i, subject in enumerate(dogtag_entries_subjects):
            entry = LDAPEntry(fake_conn,
                              DN('cn=%i,ou=certificateRepository' % i,
                                 'ou=ca,o=ipaca'),
                              userCertificate=[IPACertificate()],
                              subjectName=[subject])
            ldap_entries.append(entry)

        mock_certdb.return_value = mock_CertDB(self.trust)

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPADogtagCertsMatchCheck(registry)
        f.conn = mock_ldap(ldap_entries)
        self.results = capture_results(f)

        assert len(self.results) == 3
        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPADogtagCertsMatchCheck'

    @patch('ipaserver.install.certs.CertDB')
    def test_certs_match_ok_subject(self, mock_certdb):
        """ Ensure match check is ok"""
        m_api.Command.config_show.side_effect = custom_subject_base
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        pkidbentry = LDAPEntry(fake_conn,
                               DN('uid=pkidbuser,ou=people,o=ipaca'),
                               userCertificate=[IPACertificate()],
                               subjectName=['test'])
        casignentry = LDAPEntry(fake_conn,
                                DN('cn=%s IPA CA' % m_api.env.realm,
                                   'cn=certificates,cn=ipa,cn=etc',
                                    m_api.env.basedn),
                                CACertificate=[IPACertificate()],
                                userCertificate=[IPACertificate()],
                                subjectName=['test'])
        ldap_entries = [pkidbentry, casignentry]

        dogtag_entries_subjects = self.get_dogtag_subjects(
            m_api.env.host, custom_subject_base
        )

        for i, subject in enumerate(dogtag_entries_subjects):
            entry = LDAPEntry(fake_conn,
                              DN('cn=%i,ou=certificateRepository' % i,
                                 'ou=ca,o=ipaca'),
                              userCertificate=[IPACertificate()],
                              subjectName=[subject])
            ldap_entries.append(entry)

        mock_certdb.return_value = mock_CertDB(self.trust)

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPADogtagCertsMatchCheck(registry)
        f.conn = mock_ldap(ldap_entries)
        self.results = capture_results(f)

        assert len(self.results) == 3
        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPADogtagCertsMatchCheck'

    @patch('ipaserver.install.certs.CertDB')
    def test_certs_mismatch_subject(self, mock_certdb):
        """ Ensure mismatches are detected"""
        m_api.Command.config_show.side_effect = custom_subject_base
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        pkidbentry = LDAPEntry(fake_conn,
                               DN('uid=pkidbuser,ou=people,o=ipaca'),
                               userCertificate=[IPACertificate(
                                   serial_number=2
                               )],
                               subjectName=['test'])
        casignentry = LDAPEntry(fake_conn,
                                DN('cn=%s IPA CA' % m_api.env.realm,
                                   'cn=certificates,cn=ipa,cn=etc',
                                    m_api.env.basedn),
                                CACertificate=[IPACertificate()],
                                userCertificate=[IPACertificate()],
                                subjectName=['test'])
        ldap_entries = [pkidbentry, casignentry]

        dogtag_entries_subjects = self.get_dogtag_subjects(
            m_api.env.host, custom_subject_base
        )

        for i, subject in enumerate(dogtag_entries_subjects):
            entry = LDAPEntry(fake_conn,
                              DN('cn=%i,ou=certificateRepository' % i,
                                 'ou=ca,o=ipaca'),
                              userCertificate=[IPACertificate()],
                              subjectName=[subject])
            ldap_entries.append(entry)

        mock_certdb.return_value = mock_CertDB(self.trust)

        framework = object()
        registry.initialize(framework, config.Config())
        f = IPADogtagCertsMatchCheck(registry)
        f.conn = mock_ldap(ldap_entries)
        self.results = capture_results(f)

        assert len(self.results) == 3
        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.certs'



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#
from __future__ import division

import os
import shutil

from ipahealthcheck.system.plugin import SystemPlugin, registry
from ipahealthcheck.core.plugin import duration, Result
from ipahealthcheck.core import constants


def in_container():
    """Determine if we're running in a container."""
    with open('/proc/1/sched', 'r') as sched:
        data_sched = sched.readline()

    with open('/proc/self/cgroup', 'r') as cgroup:
        data_cgroup = cgroup.readline()

    checks = [
        data_sched.split()[0] not in ('systemd', 'init',),
        data_cgroup.split()[0] not in ('libpod'),
        os.path.exists('/.dockerenv'),
        os.path.exists('/.dockerinit'),
        os.getenv('container', None) is not None
    ]

    return any(checks)


@registry
class FileSystemSpaceCheck(SystemPlugin):
    """Check for filesystem available space."""

    # watch important directories for FreeIPA
    _pathchecks = {
        '/var/lib/dirsrv/': 1024,
        '/var/lib/ipa/backup/': 512,
        '/var/log/': 1024,
        '/var/tmp/': 512,
        '/tmp': 512
    }

    if not in_container():
        _pathchecks['/var/log/audit/'] = 512

    # File systems reaching 90% capacity risk fragmentation.
    # Defragmentation is never desirable and not available
    # on ext4 anyway. So error out at 20% free space.
    min_free_percent = 20

    def get_fs_free_space(self, pathname):
        stat = shutil.disk_usage(pathname)
        return int(stat.free / 2**20)

    def get_fs_free_space_percentage(self, pathname):
        stat = shutil.disk_usage(pathname)
        return int(stat.free * 100 / stat.total)

    @duration
    def check(self):
        for store, threshold in self._pathchecks.items():
            try:
                percent_free = self.get_fs_free_space_percentage(store)
            except FileNotFoundError:
                yield Result(
                    self, constants.WARNING,
                    key=store,
                    msg='File system {store} is not mounted',
                    store=store
                )
                continue
            if percent_free < self.min_free_percent:
                yield Result(
                    self, constants.ERROR,
                    key=store,
                    msg='%s: %s %s%% < %s%%' % (
                        store, 'free space percentage under threshold:',
                        percent_free, self.min_free_percent
                    ),
                    store=store, percent_free=percent_free,
                    threshold=self.min_free_percent
                )
            else:
                yield Result(
                    self, constants.SUCCESS,
                    key=store,
                    msg='%s: %s %s%% >= %s%%' % (
                        store, 'free space percentage within limits:',
                        percent_free, self.min_free_percent
                    ),
                    store=store, percent_free=percent_free,
                    threshold=self.min_free_percent
                )
            free_space = self.get_fs_free_space(store)
            if free_space < threshold:
                yield Result(
                    self, constants.ERROR,
                    key=store,
                    msg='%s: %s %s MiB < %s MiB' % (
                        store, 'free space under threshold:',
                        free_space, threshold
                    ),
                    store=store, free_space=free_space, threshold=threshold
                )
            else:
                yield Result(
                    self, constants.SUCCESS,
                    key=store,
                    msg='%s: %s %s MiB >= %s MiB' % (
                        store, 'free space within limits:',
                        free_space, threshold
                    ),
                    store=store, free_space=free_space, threshold=threshold
                )



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from base import BaseTest
from unittest.mock import patch
from util import capture_results, CAInstance
from util import m_api

from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.roles import (IPACRLManagerCheck,
                                      IPARenewalMasterCheck)


class TestCRLManagerRole(BaseTest):
    @patch('ipaserver.install.cainstance.CAInstance')
    def test_not_crlmanager(self, mock_ca):
        mock_ca.return_value = CAInstance(crlgen=False)
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPACRLManagerCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.roles'
        assert result.check == 'IPACRLManagerCheck'
        assert result.kw.get('crlgen_enabled') is False

    @patch('ipaserver.install.cainstance.CAInstance')
    def test_crlmanager(self, mock_ca):
        mock_ca.return_value = CAInstance()
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPACRLManagerCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.roles'
        assert result.check == 'IPACRLManagerCheck'
        assert result.kw.get('crlgen_enabled') is True

    @patch('ipaserver.install.cainstance.CAInstance')
    def test_crlmanager_no_ca(self, mock_ca):
        """There should be no CRLManagerCheck without a CA"""
        mock_ca.return_value = CAInstance(False)
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPACRLManagerCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 0


class TestRenewalMaster(BaseTest):
    def test_renewal_master_not_set(self):
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPARenewalMasterCheck(registry)

        m_api.Command.config_show.side_effect = [{
            'result': {
            }
        }]

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.roles'
        assert result.check == 'IPARenewalMasterCheck'
        assert result.kw.get('master') is False

    def test_not_renewal_master(self):
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPARenewalMasterCheck(registry)

        m_api.Command.config_show.side_effect = [{
            'result': {
                'ca_renewal_master_server': 'something.ipa.example'
            }
        }]

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.roles'
        assert result.check == 'IPARenewalMasterCheck'
        assert result.kw.get('master') is False

    def test_is_renewal_master(self):
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPARenewalMasterCheck(registry)

        m_api.Command.config_show.side_effect = [{
            'result': {
                'ca_renewal_master_server': 'server.ipa.example'
            }
        }]

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.roles'
        assert result.check == 'IPARenewalMasterCheck'
        assert result.kw.get('master') is True



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.core.plugin import Results
from unittest.mock import patch, Mock
import ipalib
from ipapython.dn import DN


class ExceptionNotRaised(Exception):
    """
    Exception raised when an *expected* exception is *not* raised during a
    unit test.
    """
    msg = 'expected %s'

    def __init__(self, expected):
        self.expected = expected

    def __str__(self):
        return self.msg % self.expected.__name__


def raises(exception, callback, *args, **kw):
    """
    Tests that the expected exception is raised; raises ExceptionNotRaised
    if test fails.
    """
    try:
        callback(*args, **kw)
    except exception as e:
        return e
    raise ExceptionNotRaised(exception)


def capture_results(f):
    """
    Loop over check() and collect the results.
    """
    results = Results()
    for result in f.check():
        if result is not None:
            results.add(result)

    return results


class DogtagInstance:
    def __init__(self, hsm_enabled=False, token_name=None):
        self.hsmenabled = hsm_enabled
        self.tokenname = token_name

    @property
    def hsm_enabled(self):
        """Is HSM support enabled?"""
        return self.hsmenabled

    @property
    def token_name(self):
        """HSM token name"""
        return self.tokenname


class CAInstance(DogtagInstance):
    """A bare-bones CAinistance override

       This is needed to control whether the underlying master is
       CAless or CAful.
    """
    def __init__(self, enabled=True, crlgen=True, hsm_enabled=False,
                 token=None):
        super(CAInstance, self).__init__(hsm_enabled=hsm_enabled,
                                         token_name=token)
        self.enabled = enabled
        self.crlgen = crlgen

    def is_configured(self):
        return self.enabled

    def is_crlgen_enabled(self):
        return self.crlgen


class KRAInstance(DogtagInstance):
    """A bare-bones KRAinistance override

       This is needed to control whether the underlying master is
       has a KRA installed or not.
    """
    def __init__(self, installed=True):
        self.installed = installed

    def is_installed(self):
        return self.installed


class ServiceBasedRole:
    """A bare-bones role override

       This is just enough to satisfy the initialization code so
       the AD Trust status can be determined. It will always default
       to false and the registry should be overridden directly in the
       test cases.
    """
    def __init__(self, attr_name=None, name=None, component_services=None):
        pass

    def status(self, api_instance, server=None, attrs_list=("*",)):
        return [dict()]


class ADtrustBasedRole(ServiceBasedRole):
    """A bare-bones role override

       This is just enough to satisfy the initialization code so
       the AD Trust status can be determined. It will always default
       to false and the registry should be overridden directly in the
       test cases.
    """
    def __init__(self, attr_name=None, name=None):
        pass


# Mock api. This file needs to be imported before anything that would
# import ipalib.api in order for it to be replaced properly.

p_api = patch('ipalib.api', autospec=ipalib.api)
m_api = p_api.start()
m_api.isdone.return_value = True
m_api.env = Mock()
m_api.env.host = 'server.ipa.example'
m_api.env.server = 'server.ipa.example'
m_api.env.realm = u'IPA.EXAMPLE'
m_api.env.domain = u'ipa.example'
m_api.env.basedn = u'dc=ipa,dc=example'
m_api.env.container_user = DN(('cn', 'users'), ('cn', 'accounts'))
m_api.env.container_group = DN(('cn', 'groups'), ('cn', 'accounts'))
m_api.env.container_host = DN(('cn', 'computers'), ('cn', 'accounts'))
m_api.env.container_sysaccounts = DN(('cn', 'sysaccounts'), ('cn', 'etc'))
m_api.env.container_service = DN(('cn', 'services'), ('cn', 'accounts'))
m_api.env.container_masters = DN(('cn', 'masters'))
m_api.Backend = Mock()
m_api.Command = Mock()
m_api.Command.ping.return_value = {
    u'summary': u'IPA server version 4.4.3. API version 2.215',
}


def no_exceptions(results):
    """Given Results ensure that an except was not raised"""
    for result in results.results:
        assert 'exception' not in result.kw



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import logging
import os
import socket
from ipahealthcheck.core import constants
from ipahealthcheck.core.exceptions import TimeoutError
from ipahealthcheck.core.plugin import Result, duration
from ipahealthcheck.meta.plugin import Plugin, registry
from ipapython import ipautil
from ipapython.version import VERSION, API_VERSION
from ipaplatform.paths import paths

if 'FIPS_MODE_SETUP' not in dir(paths):
    paths.FIPS_MODE_SETUP = '/usr/bin/fips-mode-setup'

logger = logging.getLogger()


@registry
class MetaCheck(Plugin):
    @duration
    def check(self):

        rval = constants.SUCCESS
        if not os.path.exists(paths.FIPS_MODE_SETUP):
            fips = "missing {}".format(paths.FIPS_MODE_SETUP)
            logger.debug('%s is not installed, skipping',
                         paths.FIPS_MODE_SETUP)
        else:
            try:
                result = ipautil.run([paths.FIPS_MODE_SETUP,
                                      '--is-enabled'],
                                     capture_output=True,
                                     raiseonerr=False,)
            except TimeoutError:
                logger.debug('fips-mode-setup timed out')
                fips = "check timed out"
                rval = constants.ERROR
            except Exception as e:
                logger.debug('fips-mode-setup failed: %s', e)
                fips = "failed to check"
                rval = constants.ERROR
            else:
                logger.debug(result.raw_output.decode('utf-8'))
                if result.returncode == 0:
                    fips = "enabled"
                elif result.returncode == 1:
                    fips = "inconsistent"
                elif result.returncode == 2:
                    fips = "disabled"
                else:
                    fips = "unknown"

        if not os.path.exists('/usr/sbin/ipa-acme-manage'):
            acme = "missing {}".format('/usr/sbin/ipa-acme-manage')
            logger.debug('%s is not installed, skipping',
                         '/usr/sbin/ipa-acme-manage')
        else:
            try:
                result = ipautil.run(['ipa-acme-manage', 'status'],
                                     capture_output=True,
                                     raiseonerr=False,)
            except TimeoutError:
                logger.debug('ipa-acme-manage timed out')
                acme = "check timed out"
                rval = constants.ERROR
            except Exception as e:
                logger.debug('ipa-acme-manage failed: %s', e)
                acme = "failed to check"
                rval = constants.ERROR
            else:
                logger.debug(result.raw_output.decode('utf-8'))
                if "disabled" in result.output_log:
                    acme = "disabled"
                elif "enabled" in result.output_log:
                    acme = "enabled"
                else:
                    acme = "unknown"

        yield Result(self, rval,
                     key='meta',
                     fqdn=socket.getfqdn(),
                     fips=fips,
                     acme=acme,
                     ipa_version=VERSION,
                     ipa_api_version=API_VERSION,)



#
# Copyright (C) 2020 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.ds.plugin import DSPlugin, registry
from lib389.replica import Replica, Changelog5


@registry
class ReplicationCheck(DSPlugin):
    """
    Check the agreement status for various states, and check for conflicts
    """
    check_class = Replica


@registry
class ReplicationChangelogCheck(DSPlugin):
    """
    Check the replication changelog has some sort of trimming configured
    """
    check_class = Changelog5




# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import logging

from ipahealthcheck.ipa.plugin import IPAPlugin, registry
from ipahealthcheck.core.plugin import Result, duration
from ipahealthcheck.core import constants

from ipalib import api
from ipaserver.install import replication


logger = logging.getLogger()


@registry
class IPADNARangeCheck(IPAPlugin):
    """
    Report the configured DNA range, if any.

    This expects some external system to analyze and determine if
    any or all masters have a DNA range configured. It is not an error
    if a master does not have a range. It IS an error if no masters have
    a range.
    """
    requires = ('dirsrv',)

    @duration
    def check(self):
        agmt = replication.ReplicationManager(api.env.realm, api.env.host)

        (range_start, range_max) = agmt.get_DNA_range(api.env.host)
        (next_start, next_max) = agmt.get_DNA_next_range(api.env.host)

        if range_start is not None:
            yield Result(self, constants.SUCCESS,
                         range_start=range_start,
                         range_max=range_max,
                         next_start=next_start or 0,
                         next_max=next_max or 0)
        else:
            yield Result(self, constants.WARNING,
                         key='no_dna_range_defined',
                         range_start=0,
                         range_max=0,
                         next_start=0,
                         next_max=0,
                         msg='No DNA range defined. If no masters define a '
                             'range then users and groups cannot be '
                             'created.')



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from util import capture_results, CAInstance
from util import m_api
from base import BaseTest
from unittest.mock import Mock

from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.certs import IPACertRevocation
from mock_certmonger import create_mock_dbus, _certmonger
from mock_certmonger import get_expected_requests, set_requests


class IPACertificate:
    def __init__(self, serial_number=1):
        self.subject = 'CN=RA AGENT'
        self.issuer = 'CN=ISSUER'
        self.serial_number = serial_number


class TestRevocation(BaseTest):
    patches = {
        'ipaserver.install.certs.is_ipa_issued_cert':
        Mock(return_value=True),
        'ipalib.x509.load_certificate_from_file':
        Mock(return_value=IPACertificate()),
        'ipahealthcheck.ipa.certs.get_expected_requests':
        Mock(return_value=get_expected_requests()),
        'ipalib.install.certmonger._cm_dbus_object':
        Mock(side_effect=create_mock_dbus),
        'ipalib.install.certmonger._certmonger':
        Mock(return_value=_certmonger()),
        'ipaserver.install.cainstance.CAInstance':
        Mock(return_value=CAInstance()),
    }

    def test_revocation_ok(self):
        m_api.Command.cert_show.side_effect = [
            {
                u'result': {
                    u"revoked": False,
                }
            },
            {
                u'result': {
                    u"revoked": False,
                }
            },
        ]

        set_requests()

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPACertRevocation(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.certs'
            assert result.check == 'IPACertRevocation'

    def test_revocation_one_bad(self):
        m_api.Command.cert_show.side_effect = [
            {
                u'result': {
                    u"revoked": False,
                }
            },
            {
                u'result': {
                    u"revoked": True,
                    u"revocation_reason": 4,
                }
            },
        ]
        set_requests()

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPACertRevocation(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPACertRevocation'

        result = self.results.results[1]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.certs'
        assert result.check == 'IPACertRevocation'
        assert result.kw.get('revocation_reason') == 'superseded'



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import sys

from base import BaseTest
from collections import namedtuple
from unittest.mock import Mock, patch
from util import capture_results
from util import m_api

from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.trust import (IPATrustAgentCheck,
                                      IPATrustDomainsCheck,
                                      IPADomainCheck,
                                      IPATrustCatalogCheck,
                                      IPAsidgenpluginCheck,
                                      IPATrustAgentMemberCheck,
                                      IPATrustControllerPrincipalCheck,
                                      IPATrustControllerServiceCheck,
                                      IPATrustControllerGroupSIDCheck,
                                      IPATrustControllerAdminSIDCheck,
                                      IPATrustControllerConfCheck,
                                      IPATrustPackageCheck)

from ipalib import errors
from ipapython.dn import DN
from ipapython.ipaldap import LDAPClient, LDAPEntry
try:
    from ipaserver.masters import ENABLED_SERVICE, HIDDEN_SERVICE
except ImportError:
    from ipaserver.install.service import ENABLED_SERVICE, HIDDEN_SERVICE


try:
    from ipapython.ipaldap import realm_to_serverid
except ImportError:
    from ipaserver.install.installutils import realm_to_serverid

from ldap import OPT_X_SASL_SSF_MIN
from SSSDConfig import NoOptionError


class mock_ldap:
    SCOPE_BASE = 1
    SCOPE_ONELEVEL = 2
    SCOPE_SUBTREE = 4

    def __init__(self, ldapentry):
        """Initialize the results that we will return from get_entries"""
        self.results = ldapentry

#    def get_entries(self, base_dn, scope=SCOPE_SUBTREE, filter=None,
#                    attrs_list=None, get_effective_rights=False, **kwargs):
#        if self.results is None:
#            raise errors.NotFound(reason='test')
#        return self.results
    def get_entry(self, dn, attrs_list=None, time_limit=None,
                  size_limit=None, get_effective_rights=False):
        if self.results is None:
            raise errors.NotFound(reason='test')
        return self.results


class mock_ldap_conn:
    def set_option(self, option, invalue):
        pass

    def get_option(self, option):
        if option == OPT_X_SASL_SSF_MIN:
            return 256

        return None

    def search_s(self, base, scope, filterstr=None,
                 attrlist=None, attrsonly=0):
        return tuple()


#
# Construct a setup with two direct trusts and one sub domain
#
def trust_find():
    return [{
        'result': [
            {
                'cn': ['ad.example'],
            },
            {
                'cn': ['child.example'],
            },
        ]
    }]


def trustdomain_find():
    return [
        {
            "result": [
                {
                    "cn": ["ad.example"],
                    "ipantflatname": ["ADROOT"],
                    "ipanttrusteddomainsid": ["S-1-5-21-abc"],
                    "ipanttrusttype": ["2"],
                    "ipanttrustattributes": ["8"],
                },
                {
                    "cn": ["child.ad.example"],
                    "ipantflatname": ["CHILD.ADROOT"],
                    "ipanttrusteddomainsid": ["S-1-5-22-def"],
                    "ipanttrusttype": ["2"],
                    "ipanttrustattributes": ["1"],
                },
            ],
        },
        {
            "result": [
                {
                    "cn": ["child.example"],
                    "ipantflatname": ["CHILD"],
                    "ipanttrusteddomainsid": ["S-1-5-21-ghi"],
                    "ipanttrusttype": ["2"],
                    "ipanttrustattributes": ["8"],
                },
            ],
        },
    ]


class SSSDDomain:
    def __init__(self, return_ipa_server_mode=True, provider='ipa'):
        self.return_ipa_server_mode = return_ipa_server_mode
        self.provider = provider

    def get_option(self, option):
        if option in ('id_provider', 'auth_provider', 'chpass_provider',
                      'access_provider'):
            return self.provider
        if option == 'ipa_server_mode':
            if self.return_ipa_server_mode is None:
                raise NoOptionError()
            return self.return_ipa_server_mode

        return None


class SSSDConfig():
    def __init__(self, return_domains=True, return_ipa_server_mode=True,
                 provider='ipa'):
        """
        Knobs to control what data the configuration returns.
        """
        self.return_domains = return_domains
        self.return_ipa_server_mode = return_ipa_server_mode
        self.provider = provider

    def import_config(self):
        pass

    def list_active_domains(self):
        return ('ipa.example',)

    def get_domain(self, name):
        return SSSDDomain(self.return_ipa_server_mode, self.provider)


class TestTrustAgent(BaseTest):
    def test_no_trust_agent(self):
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = False
        f = IPATrustAgentCheck(registry)

        self.results = capture_results(f)

        # Zero because the call was skipped altogether
        assert len(self.results) == 0

    @patch('SSSDConfig.SSSDConfig')
    def test_trust_agent_ok(self, mock_sssd):
        mock_sssd.return_value = SSSDConfig(return_domains=True,
                                            return_ipa_server_mode=True)
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPATrustAgentCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustAgentCheck'

    @patch('SSSDConfig.SSSDConfig')
    def test_trust_agent_not_ipa(self, mock_sssd):
        mock_sssd.return_value = SSSDConfig(return_domains=True,
                                            return_ipa_server_mode=False)
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPATrustAgentCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustAgentCheck'
        assert result.kw.get('key') == 'ipa_server_mode_false'
        assert result.kw.get('domain') == 'ipa.example'

    @patch('SSSDConfig.SSSDConfig')
    def test_trust_agent_fail(self, mock_sssd):
        mock_sssd.return_value = SSSDConfig(return_domains=True,
                                            return_ipa_server_mode=None)
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPATrustAgentCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustAgentCheck'
        assert result.kw.get('key') == 'ipa_server_mode_missing'
        assert result.kw.get('domain') == 'ipa.example'


class TestTrustDomains(BaseTest):
    def test_no_trust_agent(self):
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = False
        f = IPATrustDomainsCheck(registry)

        self.results = capture_results(f)

        # Zero because the call was skipped altogether
        assert len(self.results) == 0

    @patch('ipapython.ipautil.run')
    def test_trust_domain_list_fail(self, mock_run):
        run_result = namedtuple('run', ['returncode', 'error_log'])
        run_result.returncode = 1
        run_result.error_log = 'error'
        mock_run.return_value = run_result

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPATrustDomainsCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustDomainsCheck'
        assert result.kw.get('key') == 'domain_list_error'

    @patch('ipapython.ipautil.run')
    @patch('ipahealthcheck.ipa.trust.get_trust_domains')
    def test_trust_get_trust_domains_fail(self, mock_trust, mock_run):
        # sssctl domain-list
        run_result = namedtuple('run', ['returncode', 'error_log'])
        run_result.returncode = 0
        run_result.error_log = ''
        run_result.output = 'implicit_files\nipa.example\nad.example\n'
        mock_run.return_value = run_result

        mock_trust.side_effect = errors.NotFound(reason='bad')

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPATrustDomainsCheck(registry)

        self.results = capture_results(f)

        # There are more than one result I just care about this particular
        # value. The error is not fatal.
        result = self.results.results[0]
        assert result.result == constants.WARNING
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustDomainsCheck'
        assert result.kw.get('key') == 'trust-find'

    @patch('ipapython.ipautil.run')
    def test_trust_get_trust_domains_ok(self, mock_run):
        # sssctl domain-list
        dlresult = namedtuple('run', ['returncode', 'error_log'])
        dlresult.returncode = 0
        dlresult.error_log = ''
        dlresult.output = 'implicit_files\nipa.example\nad.example\n' \
            'child.ad.example\nchild.example\n'
        olresult = namedtuple('run', ['returncode', 'error_log'])
        olresult.returncode = 0
        olresult.error_log = ''
        olresult.output = 'Online status: Online\n\n'

        mock_run.side_effect = [dlresult, olresult, olresult, olresult]

        # get_trust_domains()
        m_api.Command.trust_find.side_effect = trust_find()
        m_api.Command.trustdomain_find.side_effect = trustdomain_find()

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPATrustDomainsCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 4

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustDomainsCheck'
        assert result.kw.get('key') == 'domain-list'
        assert result.kw.get('trust_domains') == \
            'ad.example, child.ad.example, child.example'
        assert result.kw.get('sssd_domains') == \
            'ad.example, child.ad.example, child.example'

        result = self.results.results[1]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustDomainsCheck'
        assert result.kw.get('key') == 'domain-status'
        assert result.kw.get('domain') == 'ad.example'

        result = self.results.results[2]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustDomainsCheck'
        assert result.kw.get('key') == 'domain-status'
        assert result.kw.get('domain') == 'child.ad.example'

        result = self.results.results[3]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustDomainsCheck'
        assert result.kw.get('key') == 'domain-status'
        assert result.kw.get('domain') == 'child.example'

    @patch('ipapython.ipautil.run')
    def test_trust_get_trust_domains_mismatch(self, mock_run):
        # sssctl domain-list
        dlresult = namedtuple('run', ['returncode', 'error_log'])
        dlresult.returncode = 0
        dlresult.error_log = ''
        dlresult.output = 'implicit_files\nipa.example\n' \
            'child.example\n'
        olresult = namedtuple('run', ['returncode', 'error_log'])
        olresult.returncode = 0
        olresult.error_log = ''
        olresult.output = 'Online status: Online\n\n'

        mock_run.side_effect = [dlresult, olresult, olresult]

        # get_trust_domains()
        m_api.Command.trust_find.side_effect = trust_find()
        m_api.Command.trustdomain_find.side_effect = trustdomain_find()

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPATrustDomainsCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 2

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustDomainsCheck'
        assert result.kw.get('key') == 'domain-list'
        assert result.kw.get('trust_domains') == \
            'ad.example, child.ad.example, child.example'
        assert result.kw.get('sssd_domains') == 'child.example'

        result = self.results.results[1]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustDomainsCheck'
        assert result.kw.get('key') == 'domain-status'
        assert result.kw.get('domain') == 'child.example'


class TestIPADomain(BaseTest):
    @patch('SSSDConfig.SSSDConfig')
    def test_ipa_domain_ok(self, mock_sssd):
        mock_sssd.return_value = SSSDConfig(provider='ipa')
        framework = object()
        registry.initialize(framework, config.Config)
        # being a trust agent isn't mandatory, test without
        registry.trust_agent = False
        f = IPADomainCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPADomainCheck'

    @patch('SSSDConfig.SSSDConfig')
    def test_ipa_domain_ad(self, mock_sssd):
        mock_sssd.return_value = SSSDConfig(provider='ad')
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPADomainCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 4

        for result in self.results.results:
            assert result.result == constants.ERROR
            assert result.source == 'ipahealthcheck.ipa.trust'
            assert result.check == 'IPADomainCheck'
            assert result.kw.get('provider') == 'ad'


class TestTrustCatalog(BaseTest):
    def test_no_trust_agent(self):
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = False
        f = IPATrustCatalogCheck(registry)

        self.results = capture_results(f)

        # Zero because the call was skipped altogether
        assert len(self.results) == 0

    @patch('pysss_nss_idmap.getnamebysid')
    @patch('ipapython.ipautil.run')
    def test_trust_catalog_ok(self, mock_run, mock_getnamebysid):
        # id Administrator@ad.example
        dsresult = namedtuple('run', ['returncode', 'error_log'])
        dsresult.returncode = 0
        dsresult.error_log = ''
        dsresult.output = 'Active servers:\nAD Global Catalog: ' \
            'root-dc.ad.vm\nAD Domain Controller: root-dc.ad.vm\n' \
            'IPA: master.ipa.vm\n\n'
        ds2result = namedtuple('run', ['returncode', 'error_log'])
        ds2result.returncode = 0
        ds2result.error_log = ''
        ds2result.output = 'Active servers:\nAD Global Catalog: ' \
            'root-dc.ad.vm\nAD Domain Controller: root-dc.ad.vm\n' \

        mock_run.side_effect = [dsresult, dsresult, ds2result]
        mock_getnamebysid.side_effect = [
           {'S-1-5-21-abc-500': {'name': 'admin@ad.example', 'type': 3}},
           {'S-1-5-21-ghi-500': {'name': 'admin@child.ad.example', 'type': 3}},
           {'S-1-5-21-def-500': {'name': 'admin@child.example', 'type': 3}}
        ]

        # get_trust_domains()
        m_api.Command.trust_find.side_effect = trust_find()
        m_api.Command.trustdomain_find.side_effect = trustdomain_find()

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPATrustCatalogCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 9

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustCatalogCheck'
        assert result.kw.get('key') == 'Domain Security Identifier'
        assert result.kw.get('sid') == 'S-1-5-21-abc'

        result = self.results.results[1]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustCatalogCheck'
        assert result.kw.get('key') == 'AD Global Catalog'
        assert result.kw.get('domain') == 'ad.example'

        result = self.results.results[2]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustCatalogCheck'
        assert result.kw.get('key') == 'AD Domain Controller'
        assert result.kw.get('domain') == 'ad.example'

        result = self.results.results[3]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustCatalogCheck'
        assert result.kw.get('key') == 'Domain Security Identifier'
        assert result.kw.get('sid') == 'S-1-5-22-def'

        result = self.results.results[4]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustCatalogCheck'
        assert result.kw.get('key') == 'AD Global Catalog'
        assert result.kw.get('domain') == 'child.ad.example'

        result = self.results.results[5]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustCatalogCheck'
        assert result.kw.get('key') == 'AD Domain Controller'

        result = self.results.results[6]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustCatalogCheck'
        assert result.kw.get('key') == 'Domain Security Identifier'
        assert result.kw.get('sid') == 'S-1-5-21-ghi'

        result = self.results.results[7]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustCatalogCheck'
        assert result.kw.get('key') == 'AD Global Catalog'
        assert result.kw.get('domain') == 'child.example'

        result = self.results.results[8]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustCatalogCheck'
        assert result.kw.get('key') == 'AD Domain Controller'
        assert result.kw.get('domain') == 'child.example'


class Testsidgen(BaseTest):
    patches = {
        'ldap.initialize':
        Mock(return_value=mock_ldap_conn()),
    }

    def test_no_trust_agent(self):
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = False
        f = IPAsidgenpluginCheck(registry)

        self.results = capture_results(f)

        # Zero because the call was skipped altogether
        assert len(self.results) == 0

    def test_sidgen_ok(self):
        attrs = {
            'nsslapd-pluginEnabled': ['on'],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, DN('cn=plugin, cn=config'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPAsidgenpluginCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 2

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPAsidgenpluginCheck'
        assert result.kw.get('key') == 'IPA SIDGEN'

        result = self.results.results[1]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPAsidgenpluginCheck'
        assert result.kw.get('key') == 'ipa-sidgen-task'

    def test_sidgen_fail(self):
        attrs = {
            'nsslapd-pluginEnabled': ['off'],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, DN('cn=plugin, cn=config'))
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPAsidgenpluginCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 2

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPAsidgenpluginCheck'
        assert result.kw.get('key') == 'IPA SIDGEN'

        result = self.results.results[1]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPAsidgenpluginCheck'
        assert result.kw.get('key') == 'ipa-sidgen-task'


class TestTrustAgentMember(BaseTest):
    patches = {
        'ldap.initialize':
        Mock(return_value=mock_ldap_conn()),
    }

    def test_no_trust_agent(self):
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = False
        f = IPATrustAgentMemberCheck(registry)

        self.results = capture_results(f)

        # Zero because the call was skipped altogether
        assert len(self.results) == 0

    def test_member_ok(self):
        agent_dn = DN(('fqdn', m_api.env.host), m_api.env.container_host,
                      m_api.env.basedn)
        group_dn = DN(('cn', 'adtrust agents'),
                      m_api.env.container_sysaccounts,
                      m_api.env.basedn)
        attrs = {
            'memberof': [group_dn],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, agent_dn)
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPATrustAgentMemberCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustAgentMemberCheck'
        assert result.kw.get('key') == m_api.env.host

    def test_member_fail(self):
        agent_dn = DN(('fqdn', m_api.env.host), m_api.env.container_host,
                      m_api.env.basedn)
        attrs = {
            'memberof': [agent_dn],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, agent_dn)
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_agent = True
        f = IPATrustAgentMemberCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustAgentMemberCheck'
        assert result.kw.get('key') == m_api.env.host


class TestControllerPrincipal(BaseTest):
    patches = {
        'ldap.initialize':
        Mock(return_value=mock_ldap_conn()),
    }

    def test_not_trust_controller(self):
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = False
        f = IPATrustControllerPrincipalCheck(registry)

        self.results = capture_results(f)

        # Zero because the call was skipped altogether
        assert len(self.results) == 0

    def test_principal_ok(self):
        agent_dn = DN(('krbprincipalname',
                      'cifs/%s@%s' % (m_api.env.host, m_api.env.realm)),
                      m_api.env.container_service, m_api.env.basedn)
        group_dn = DN(('cn', 'adtrust agents'),
                      m_api.env.container_sysaccounts,
                      m_api.env.basedn)
        attrs = {
            'memberof': [group_dn],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, agent_dn)
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = True
        f = IPATrustControllerPrincipalCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustControllerPrincipalCheck'
        assert result.kw.get('key') == 'cifs/%s@%s' % \
                                       (m_api.env.host, m_api.env.realm)

    def test_member_fail(self):
        agent_dn = DN(('fqdn', m_api.env.host), m_api.env.container_host,
                      m_api.env.basedn)
        attrs = {
            'memberof': [agent_dn],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, agent_dn)
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = True
        f = IPATrustControllerPrincipalCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.kw.get('key') == 'cifs/%s@%s' % \
                                       (m_api.env.host, m_api.env.realm)


class TestControllerService(BaseTest):
    patches = {
        'ldap.initialize':
        Mock(return_value=mock_ldap_conn()),
    }

    def test_not_trust_controller(self):
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = False
        f = IPATrustControllerServiceCheck(registry)

        self.results = capture_results(f)

        # Zero because the call was skipped altogether
        assert len(self.results) == 0

    def test_service_enabled(self):
        service_dn = DN(('cn', 'ADTRUST'))
        for type in [ENABLED_SERVICE, HIDDEN_SERVICE]:
            attrs = {
                'ipaconfigstring': [type],
            }
            fake_conn = LDAPClient('ldap://localhost', no_schema=True)
            ldapentry = LDAPEntry(fake_conn, service_dn)
            for attr, values in attrs.items():
                ldapentry[attr] = values

            framework = object()
            registry.initialize(framework, config.Config)
            registry.trust_controller = True
            f = IPATrustControllerServiceCheck(registry)

            f.conn = mock_ldap(ldapentry)
            self.results = capture_results(f)

            assert len(self.results) == 1

            result = self.results.results[0]
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.trust'
            assert result.check == 'IPATrustControllerServiceCheck'
            assert result.kw.get('key') == 'ADTRUST'

    def test_principal_fail(self):
        service_dn = DN(('cn', 'ADTRUST'))
        attrs = {
            'ipaconfigstring': ['disabledService'],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, service_dn)
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = True
        f = IPATrustControllerServiceCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.kw.get('key') == 'ADTRUST'


class TestControllerGroupSID(BaseTest):
    patches = {
        'ldap.initialize':
        Mock(return_value=mock_ldap_conn()),
    }

    def test_not_trust_controller(self):
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = False
        f = IPATrustControllerGroupSIDCheck(registry)

        self.results = capture_results(f)

        # Zero because the call was skipped altogether
        assert len(self.results) == 0

    def test_principal_ok(self):
        admins_dn = DN(('cn', 'admins'))
        attrs = {
            'ipantsecurityidentifier':
            ['S-1-5-21-1234-5678-1976041503-512'],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, admins_dn)
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = True
        f = IPATrustControllerGroupSIDCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustControllerGroupSIDCheck'
        assert result.kw.get('key') == 'ipantsecurityidentifier'
        assert result.kw.get('rid') == 'S-1-5-21-1234-5678-1976041503-512'

    def test_principal_fail(self):
        admins_dn = DN(('cn', 'admins'))
        attrs = {
            'ipantsecurityidentifier':
            ['S-1-5-21-1234-5678-1976041503-500'],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, admins_dn)
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = True
        f = IPATrustControllerGroupSIDCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustControllerGroupSIDCheck'
        assert result.kw.get('key') == 'ipantsecurityidentifier'
        assert result.kw.get('rid') == 'S-1-5-21-1234-5678-1976041503-500'


class TestControllerAdminSID(BaseTest):
    patches = {
        'ldap.initialize':
        Mock(return_value=mock_ldap_conn()),
    }

    def test_not_trust_controller(self):
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = False
        f = IPATrustControllerAdminSIDCheck(registry)

        self.results = capture_results(f)

        # Zero because the call was skipped altogether
        assert len(self.results) == 0

    def test_principal_ok(self):
        admin_dn = DN(('uid', 'admin'))
        attrs = {
            'ipantsecurityidentifier':
            ['S-1-5-21-1234-5678-1976041503-500'],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, admin_dn)
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = True
        f = IPATrustControllerAdminSIDCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustControllerAdminSIDCheck'
        assert result.kw.get('key') == 'ipantsecurityidentifier'
        assert result.kw.get('rid') == 'S-1-5-21-1234-5678-1976041503-500'

    def test_principal_fail(self):
        admin_dn = DN(('uid', 'admin'))
        attrs = {
            'ipantsecurityidentifier':
            ['S-1-5-21-1234-5678-1976041503-400'],
        }
        fake_conn = LDAPClient('ldap://localhost', no_schema=True)
        ldapentry = LDAPEntry(fake_conn, admin_dn)
        for attr, values in attrs.items():
            ldapentry[attr] = values

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = True
        f = IPATrustControllerAdminSIDCheck(registry)

        f.conn = mock_ldap(ldapentry)
        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.ERROR
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustControllerAdminSIDCheck'
        assert result.kw.get('key') == 'ipantsecurityidentifier'
        assert result.kw.get('rid') == 'S-1-5-21-1234-5678-1976041503-400'


class TestControllerConf(BaseTest):
    patches = {
        'ldap.initialize':
        Mock(return_value=mock_ldap_conn()),
    }

    def test_not_trust_controller(self):
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = False
        f = IPATrustControllerConfCheck(registry)

        self.results = capture_results(f)

        # Zero because the call was skipped altogether
        assert len(self.results) == 0

    @patch('ipapython.ipautil.run')
    def test_ldapi_ok(self, mock_run):
        ldapi_socket = "ipasam:ldapi://%%2fvar%%2frun%%2fslapd-%s.socket" % \
                       realm_to_serverid(m_api.env.realm)
        run_result = namedtuple('run', ['returncode', 'output'])
        run_result.returncode = 0
        run_result.output = '[global]\n\tpassdb backend=%s' % ldapi_socket
        mock_run.return_value = run_result

        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = True
        f = IPATrustControllerConfCheck(registry)

        self.results = capture_results(f)

        assert len(self.results) == 1

        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustControllerConfCheck'
        assert result.kw.get('key') == 'net conf list'


class TestPackageCheck(BaseTest):
    def test_agent_with_package(self):
        # Note that this test assumes the import is installed
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = False
        registry.trust_agent = True
        f = IPATrustPackageCheck(registry)
        self.results = capture_results(f)
        assert len(self.results) == 1
        result = self.results.results[0]
        assert result.result == constants.SUCCESS
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustPackageCheck'

    def test_agent_without_package(self):
        # Note that this test assumes the import is installed
        framework = object()
        registry.initialize(framework, config.Config)
        registry.trust_controller = False
        registry.trust_agent = True
        # Hose up the module so the import fails
        save = sys.modules['ipaserver.install']
        sys.modules['ipaserver.install'] = 'foo'
        f = IPATrustPackageCheck(registry)
        self.results = capture_results(f)
        assert len(self.results) == 1
        result = self.results.results[0]
        assert result.result == constants.WARNING
        assert result.source == 'ipahealthcheck.ipa.trust'
        assert result.check == 'IPATrustPackageCheck'
        sys.modules['ipaserver.install'] = save



from setuptools import find_packages, setup


setup(
    name='ipahealthcheck',
    version='0.12',
    namespace_packages=['ipahealthcheck', 'ipaclustercheck'],
    package_dir={'': 'src'},
    # packages=find_packages(where='src'),
    packages=[
        'ipahealthcheck.core',
        'ipahealthcheck.dogtag',
        'ipahealthcheck.ds',
        'ipahealthcheck.ipa',
        'ipahealthcheck.meta',
        'ipahealthcheck.system',
        'ipaclustercheck.core',
        'ipaclustercheck.ipa',
    ],
    entry_points={
        # creates bin/ipahealthcheck
        'console_scripts': [
            'ipa-healthcheck = ipahealthcheck.core.main:main',
            'ipa-clustercheck = ipaclustercheck.core.main:main',
        ],
        # subsystem registries
        'ipahealthcheck.registry': [
            'ipahealthcheck.dogtag = ipahealthcheck.dogtag.plugin:registry',
            'ipahealthcheck.ipa = ipahealthcheck.ipa.plugin:registry',
            'ipahealthcheck.meta = ipahealthcheck.meta.plugin:registry',
            'ipahealthcheck.ds = ipahealthcheck.ds.plugin:registry',
            'ipahealthcheck.system = ipahealthcheck.system.plugin:registry'
        ],
        # plugin modules for ipahealthcheck.meta registry
        'ipahealthcheck.meta': [
            'meta = ipahealthcheck.meta.core',
            'services = ipahealthcheck.meta.services',
        ],
        # plugin modules for ipahealthcheck.ipa registry
        'ipahealthcheck.ipa': [
            'ipacerts = ipahealthcheck.ipa.certs',
            'ipadna = ipahealthcheck.ipa.dna',
            'ipadns = ipahealthcheck.ipa.idns',
            'ipafiles = ipahealthcheck.ipa.files',
            'ipahost = ipahealthcheck.ipa.host',
            'ipakdc = ipahealthcheck.ipa.kdc',
            'ipaproxy = ipahealthcheck.ipa.proxy',
            'ipameta = ipahealthcheck.ipa.meta',
            'ipanss = ipahealthcheck.ipa.nss',
            'iparoles = ipahealthcheck.ipa.roles',
            'ipatopology = ipahealthcheck.ipa.topology',
            'ipatrust = ipahealthcheck.ipa.trust',
        ],
        # plugin modules for ipahealthcheck.dogtag registry
        'ipahealthcheck.dogtag': [
            'dogtagca = ipahealthcheck.dogtag.ca',
        ],
        # plugin modules for ipahealthcheck.ds registry
        'ipahealthcheck.ds': [
            'dsbackends = ipahealthcheck.ds.backends',
            'dsconfig = ipahealthcheck.ds.config',
            'dsdiskspace = ipahealthcheck.ds.disk_space',
            'dsdse = ipahealthcheck.ds.dse',
            'dsencryption = ipahealthcheck.ds.encryption',
            'dsfschecks = ipahealthcheck.ds.fs_checks',
            'dsnssssl = ipahealthcheck.ds.nss_ssl',
            'dsplugins = ipahealthcheck.ds.ds_plugins',
            'dsreplication = ipahealthcheck.ds.replication',
            'dsruv = ipahealthcheck.ds.ruv',
        ],
        # plugin modules for ipahealthcheck.system registry
        'ipahealthcheck.system': [
            'filesystemspace = ipahealthcheck.system.filesystemspace',
        ],
        'ipaclustercheck.registry': [
            'ipaclustercheck.ipa = ipaclustercheck.ipa.plugin:registry',
        ],
        'ipaclustercheck.ipa': [
            'crl = ipaclustercheck.ipa.crlmanager',
            'ruv = ipaclustercheck.ipa.ruv',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
)



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#
from __future__ import division

from base import BaseTest
from unittest.mock import Mock
from util import capture_results
from collections import namedtuple

from ipahealthcheck.core import config, constants
from ipahealthcheck.system.plugin import registry
from ipahealthcheck.system.filesystemspace import FileSystemSpaceCheck
from ipahealthcheck.system.filesystemspace import in_container


class TestFileSystemNotEnoughFreeSpace(BaseTest):

    usage = namedtuple('usage', ['total', 'used', 'free'])
    usage.total = 2087428096
    usage.used = 1628193914
    usage.free = 459234182

    patches = {
        'shutil.disk_usage':
        Mock(return_value=usage),
    }

    def test_filesystem_near_enospc(self):

        framework = object()
        registry.initialize(framework, config.Config)
        f = FileSystemSpaceCheck(registry)

        self.results = capture_results(f)

        expected_results = 10 if in_container() else 12

        count = 0
        for result in self.results.results:
            if result.result == constants.ERROR:
                count += 1
                assert result.source == 'ipahealthcheck.system.filesystemspace'
                assert result.check == 'FileSystemSpaceCheck'
                assert 'free space under threshold' in result.kw.get('msg')
            else:
                assert 'free space percentage within' in result.kw.get('msg')
        assert len(self.results) == expected_results
        assert count == expected_results / 2


class TestFileSystemNotEnoughFreeSpacePercentage(BaseTest):

    usage = namedtuple('usage', ['total', 'used', 'free'])
    usage.total = 10437140480
    usage.used = 8913305600
    usage.free = 1523834880

    patches = {
        'shutil.disk_usage':
        Mock(return_value=usage),
    }

    def test_filesystem_risking_fragmentation(self):

        framework = object()
        registry.initialize(framework, config.Config)
        f = FileSystemSpaceCheck(registry)

        self.results = capture_results(f)

        expected_results = 10 if in_container() else 12

        count = 0
        for result in self.results.results:
            if result.result == constants.ERROR:
                count += 1
                assert result.source == 'ipahealthcheck.system.filesystemspace'
                assert result.check == 'FileSystemSpaceCheck'
                assert 'free space percentage under' in result.kw.get('msg')
            else:
                assert 'free space within limits' in result.kw.get('msg')
        assert len(self.results) == expected_results
        assert count == expected_results / 2


class TestFileSystemEnoughFreeSpace(BaseTest):

    usage = namedtuple('usage', ['total', 'used', 'free'])
    usage.total = 10437140480
    usage.used = 1523834880
    usage.free = 8913305600

    patches = {
        'shutil.disk_usage':
        Mock(return_value=usage),
    }

    def test_filesystem_with_enough_space(self):

        framework = object()
        registry.initialize(framework, config.Config)
        f = FileSystemSpaceCheck(registry)

        self.results = capture_results(f)

        expected_results = 10 if in_container() else 12

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.system.filesystemspace'
            assert result.check == 'FileSystemSpaceCheck'
            assert (
                'free space percentage within' in result.kw.get('msg') or
                'free space within limits' in result.kw.get('msg')
            )
        assert len(self.results) == expected_results



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import logging
import os
from configparser import ConfigParser, ParsingError
from collections import OrderedDict

from ipahealthcheck.core.constants import CONFIG_SECTION, EXCLUDE_SECTION
from ipahealthcheck.core.constants import DEFAULT_CONFIG

logger = logging.getLogger()


class DuplicateOrderedDict(OrderedDict):
    def __setitem__(self, key, value):
        """Duplicate keys will be concatenated strings separated by new-line"""
        if isinstance(value, list) and key in self:
            self[key].extend(value)
        else:
            super().__setitem__(key, value)


class Config:
    """Helper class to manage configuration

       Let one treat config items as properties instead of using
       a dict. It just allows for an easier-to-read shorthand.

       >>> config = Config()
       >>> config.foo = 'bar'
       >>> config.foo
       'bar'

       Return a list of the configuration option keys.

       >>> list(config)
       ['foo']
    """

    def __init__(self):
        object.__setattr__(self, '_Config__d', {})

    def __setattr__(self, key, value):
        """
        Set the attribute named ``name`` to ``value``.
        """

        self[key] = value

    def __setitem__(self, key, value):
        """
        Set ``key`` to ``value``.
        """
        object.__setattr__(self, key, value)
        self.__d[key] = value

    def __getattr__(self, key):
        """
        Return the value corresponding to ``key``.
        """
        return self.__d[key]

    def __getitem__(self, key):
        """
        Return the value corresponding to ``key``.
        """
        return self.__d[key]

    def __contains__(self, key):
        """
        Return True if instance contains ``key``; otherwise return False.
        """
        return key in self.__d

    def __iter__(self):
        """
        Iterate through keys in ascending order.
        """
        for key in sorted(self.__d):
            yield key

    def merge(self, d):
        """
        Merge variables from dict ``d`` into the configuration

        The first one wins.

        :param d: dict containing configuration
        """
        for key in d:
            self.__d[key] = d[key]


def convert_string(value):
    """
    Reading options from the configuration file will leave them as
    strings. This breaks boolean values so attempt to convert them.
    """
    if not isinstance(value, str):
        return value

    if value.lower() in (
        "true",
        "false",
    ):
        return value.lower() == 'true'
    else:
        try:
            value = int(value)
        except ValueError:
            pass
    return value


def read_config(config_file):
    """
    Simple configuration file reader

    Read and return the configuration for only the default section.

    Returns a dict on success, None on failure
    """
    config = Config()
    config.merge(DEFAULT_CONFIG)
    if not os.path.exists(config_file):
        logging.warning(
            "config file %s does not exist, using defaults", config_file
        )
        return config

    parser = ConfigParser(dict_type=DuplicateOrderedDict, strict=False,
                          delimiters='=')
    try:
        parser.read(config_file)
    except ParsingError as e:
        logging.error("Unable to parse %s: %s", config_file, e)
        return None
    if not parser.has_section(CONFIG_SECTION):
        logging.error(
            "Config file %s missing %s section", config_file, CONFIG_SECTION
        )
        return None

    items = parser.items(CONFIG_SECTION)

    for (key, value) in items:
        if not key.startswith('excludes_'):
            if len(value) == 0 or value is None:
                logging.error(
                    "Empty value for %s in %s [%s]",
                    key, config_file, CONFIG_SECTION
                )
                return None
            else:
                # Try to do some basic validation. This is unfortunately
                # hardcoded.
                if key in ('all', 'debug', 'failures_only', 'verbose'):
                    if value.lower() not in ('true', 'false'):
                        logging.error(
                            "%s is not a valid boolean in %s [%s]",
                            key, config_file, CONFIG_SECTION
                        )
                        return None
                elif key in ('indent', 'timeout'):
                    if not isinstance(convert_string(value), int):
                        logging.error(
                            "%s is not a valid integer in %s [%s]",
                            key, config_file, CONFIG_SECTION
                        )
                        return None
                # Some rough type translation from strings
                config[key] = convert_string(value)

    if parser.has_section(EXCLUDE_SECTION):
        items = parser.items(EXCLUDE_SECTION)
        for (key, value) in items:
            config[EXCLUDE_SECTION + '_' + key] = value.split(os.linesep)

    return config



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#


ONE_MASTER = {
    'ipa.ipa.example': [
        {
            "source": "ipahealthcheck.meta.core",
            "check": "MetaCheck",
            "result": "SUCCESS",
            "kw": {
              "fqdn": "ipa.ipa.example",
              "ipa_version": "4.8.4",
              "ipa_api_version": "2.235"
            }
        },
        {
          "source": "ipahealthcheck.ipa.meta",
          "check": "IPAMetaCheck",
          "result": "SUCCESS",
          "kw": {
            "masters": [
              "ipa.ipa.example",
            ]
          }
        },
        # No RUV's on a freshly installed master
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
            }
        },
    ]
}


THREE_MASTERS_OK = {
    'ipa.ipa.example': [
        {
            "source": "ipahealthcheck.meta.core",
            "check": "MetaCheck",
            "result": "SUCCESS",
            "kw": {
              "fqdn": "ipa.ipa.example",
              "ipa_version": "4.8.4",
              "ipa_api_version": "2.235"
            }
        },
        {
          "source": "ipahealthcheck.ipa.meta",
          "check": "IPAMetaCheck",
          "result": "SUCCESS",
          "kw": {
            "masters": [
              "ipa.ipa.example",
              "replica1.ipa.example",
              "replica2.ipa.example"
            ]
          }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "dc=ipa,dc=example",
              "ruv": "4"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "o=ipaca",
              "ruv": "6"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "KnownRUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "ruvs_dc=ipa,dc=example",
              "suffix": "dc=ipa,dc=example",
              "ruvs": [
                [
                  "ipa.ipa.example",
                  "4"
                ],
                [
                  "replica2.ipa.example",
                  "7"
                ],
                [
                  "replica1.ipa.example",
                  "3"
                ]
              ]
            }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_o=ipaca",
            "suffix": "o=ipaca",
            "ruvs": [
              [
                "ipa.ipa.example",
                "6"
              ],
              [
                "replica2.ipa.example",
                "8"
              ],
              [
                "replica1.ipa.example",
                "5"
              ]
            ]
          }
        }
    ],
    'replica1.ipa.example': [
        {
            "source": "ipahealthcheck.meta.core",
            "check": "MetaCheck",
            "result": "SUCCESS",
            "kw": {
              "fqdn": "replica1.ipa.example",
              "masters": [
                "ipa.ipa.example",
                "replica1.ipa.example",
                "replica2.ipa.example",
              ],
              "ipa_version": "4.8.4",
              "ipa_api_version": "2.235"
            }
        },
        {
          "source": "ipahealthcheck.ipa.meta",
          "check": "IPAMetaCheck",
          "result": "SUCCESS",
          "kw": {
            "masters": [
              "ipa.ipa.example",
              "replica1.ipa.example",
              "replica2.ipa.example"
            ]
          }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "dc=ipa,dc=example",
              "ruv": "3"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "o=ipaca",
              "ruv": "5"
            }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_dc=ipa,dc=example",
            "suffix": "dc=ipa,dc=example",
            "ruvs": [
              [
                "replica1.ipa.example",
                "3"
              ],
              [
                "ipa.ipa.example",
                "4"
              ],
              [
                "replica2.ipa.example",
                "7"
              ]
            ]
          }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_o=ipaca",
            "suffix": "o=ipaca",
            "ruvs": [
              [
                "replica1.ipa.example",
                "5"
              ],
              [
                "ipa.ipa.example",
                "6"
              ],
              [
                "replica2.ipa.example",
                "8"
              ]
            ]
          }
        }
    ],
    'replica2.ipa.example': [
        {
            "source": "ipahealthcheck.meta.core",
            "check": "MetaCheck",
            "result": "SUCCESS",
            "kw": {
              "fqdn": "replica2.ipa.example",
              "masters": [
                "ipa.ipa.example",
                "replica1.ipa.example",
                "replica2.ipa.example",
              ],
              "ipa_version": "4.8.4",
              "ipa_api_version": "2.235"
            }
        },
        {
          "source": "ipahealthcheck.ipa.meta",
          "check": "IPAMetaCheck",
          "result": "SUCCESS",
          "kw": {
            "masters": [
              "ipa.ipa.example",
              "replica1.ipa.example",
              "replica2.ipa.example"
            ]
          }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "dc=ipa,dc=example",
              "ruv": "7"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "o=ipaca",
              "ruv": "8"
            }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_dc=ipa,dc=example",
            "suffix": "dc=ipa,dc=example",
            "ruvs": [
              [
                "replica2.ipa.example",
                "7"
              ],
              [
                "ipa.ipa.example",
                "4"
              ],
              [
                "replica1.ipa.example",
                "3"
              ]
            ]
          }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_o=ipaca",
            "suffix": "o=ipaca",
            "ruvs": [
              [
                "replica2.ipa.example",
                "8"
              ],
              [
                "ipa.ipa.example",
                "6"
              ],
              [
                "replica1.ipa.example",
                "5"
              ]
            ]
          }
        }
    ]
}


#
# Same three masters but replica1 has an extra RUV value
#
THREE_MASTERS_BAD_IPA_RUV = {
    'ipa.ipa.example': [
        {
            "source": "ipahealthcheck.meta.core",
            "check": "MetaCheck",
            "result": "SUCCESS",
            "kw": {
              "fqdn": "ipa.ipa.example",
              "masters": [
                "ipa.ipa.example",
                "replica1.ipa.example",
                "replica2.ipa.example",
              ],
              "ipa_version": "4.8.4",
              "ipa_api_version": "2.235"
            }
        },
        {
          "source": "ipahealthcheck.ipa.meta",
          "check": "IPAMetaCheck",
          "result": "SUCCESS",
          "kw": {
            "masters": [
              "ipa.ipa.example",
              "replica1.ipa.example",
              "replica2.ipa.example"
            ]
          }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "dc=ipa,dc=example",
              "ruv": "4"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "o=ipaca",
              "ruv": "6"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "KnownRUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "ruvs_dc=ipa,dc=example",
              "suffix": "dc=ipa,dc=example",
              "ruvs": [
                [
                  "ipa.ipa.example",
                  "4"
                ],
                [
                  "replica2.ipa.example",
                  "7"
                ],
                [
                  "replica1.ipa.example",
                  "3"
                ],
                [
                  "replica1.ipa.example",
                  "9"
                ]
              ]
            }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_o=ipaca",
            "suffix": "o=ipaca",
            "ruvs": [
              [
                "ipa.ipa.example",
                "6"
              ],
              [
                "replica2.ipa.example",
                "8"
              ],
              [
                "replica1.ipa.example",
                "5"
              ]
            ]
          }
        }
    ],
    'replica1.ipa.example': [
        {
            "source": "ipahealthcheck.meta.core",
            "check": "MetaCheck",
            "result": "SUCCESS",
            "kw": {
              "fqdn": "replica1.ipa.example",
              "masters": [
                "ipa.ipa.example",
                "replica1.ipa.example",
                "replica2.ipa.example",
              ],
              "ipa_version": "4.8.4",
              "ipa_api_version": "2.235"
            }
        },
        {
          "source": "ipahealthcheck.ipa.meta",
          "check": "IPAMetaCheck",
          "result": "SUCCESS",
          "kw": {
            "masters": [
              "ipa.ipa.example",
              "replica1.ipa.example",
              "replica2.ipa.example"
            ]
          }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "dc=ipa,dc=example",
              "ruv": "3"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "o=ipaca",
              "ruv": "5"
            }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_dc=ipa,dc=example",
            "suffix": "dc=ipa,dc=example",
            "ruvs": [
              [
                "replica1.ipa.example",
                "3"
              ],
              [
                "ipa.ipa.example",
                "4"
              ],
              [
                "replica2.ipa.example",
                "7"
              ]
            ]
          }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_o=ipaca",
            "suffix": "o=ipaca",
            "ruvs": [
              [
                "replica1.ipa.example",
                "5"
              ],
              [
                "ipa.ipa.example",
                "6"
              ],
              [
                "replica2.ipa.example",
                "8"
              ]
            ]
          }
        }
    ],
    'replica2.ipa.example': [
        {
            "source": "ipahealthcheck.meta.core",
            "check": "MetaCheck",
            "result": "SUCCESS",
            "kw": {
              "fqdn": "replica2.ipa.example",
              "masters": [
                "ipa.ipa.example",
                "replica1.ipa.example",
                "replica2.ipa.example",
              ],
              "ipa_version": "4.8.4",
              "ipa_api_version": "2.235"
            }
        },
        {
          "source": "ipahealthcheck.ipa.meta",
          "check": "IPAMetaCheck",
          "result": "SUCCESS",
          "kw": {
            "masters": [
              "ipa.ipa.example",
              "replica1.ipa.example",
              "replica2.ipa.example"
            ]
          }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "dc=ipa,dc=example",
              "ruv": "7"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "o=ipaca",
              "ruv": "8"
            }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_dc=ipa,dc=example",
            "suffix": "dc=ipa,dc=example",
            "ruvs": [
              [
                "replica2.ipa.example",
                "7"
              ],
              [
                "ipa.ipa.example",
                "4"
              ],
              [
                "replica1.ipa.example",
                "3"
              ]
            ]
          }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_o=ipaca",
            "suffix": "o=ipaca",
            "ruvs": [
              [
                "replica2.ipa.example",
                "8"
              ],
              [
                "ipa.ipa.example",
                "6"
              ],
              [
                "replica1.ipa.example",
                "5"
              ]
            ]
          }
        }
    ]
}


#
# Same three masters but replica2 CA has an extra RUV value
#
THREE_MASTERS_BAD_CS_RUV = {
    'ipa.ipa.example': [
        {
            "source": "ipahealthcheck.meta.core",
            "check": "MetaCheck",
            "result": "SUCCESS",
            "kw": {
              "fqdn": "ipa.ipa.example",
              "masters": [
                "ipa.ipa.example",
                "replica1.ipa.example",
                "replica2.ipa.example",
              ],
              "ipa_version": "4.8.4",
              "ipa_api_version": "2.235"
            }
        },
        {
          "source": "ipahealthcheck.ipa.meta",
          "check": "IPAMetaCheck",
          "result": "SUCCESS",
          "kw": {
            "masters": [
              "ipa.ipa.example",
              "replica1.ipa.example",
              "replica2.ipa.example"
            ]
          }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "dc=ipa,dc=example",
              "ruv": "4"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "o=ipaca",
              "ruv": "6"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "KnownRUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "ruvs_dc=ipa,dc=example",
              "suffix": "dc=ipa,dc=example",
              "ruvs": [
                [
                  "ipa.ipa.example",
                  "4"
                ],
                [
                  "replica2.ipa.example",
                  "7"
                ],
                [
                  "replica1.ipa.example",
                  "3"
                ],
              ]
            }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_o=ipaca",
            "suffix": "o=ipaca",
            "ruvs": [
              [
                "ipa.ipa.example",
                "6"
              ],
              [
                "replica2.ipa.example",
                "8"
              ],
              [
                "replica1.ipa.example",
                "5"
              ]
            ]
          }
        }
    ],
    'replica1.ipa.example': [
        {
            "source": "ipahealthcheck.meta.core",
            "check": "MetaCheck",
            "result": "SUCCESS",
            "kw": {
              "fqdn": "replica1.ipa.example",
              "masters": [
                "ipa.ipa.example",
                "replica1.ipa.example",
                "replica2.ipa.example",
              ],
              "ipa_version": "4.8.4",
              "ipa_api_version": "2.235"
            }
        },
        {
          "source": "ipahealthcheck.ipa.meta",
          "check": "IPAMetaCheck",
          "result": "SUCCESS",
          "kw": {
            "masters": [
              "ipa.ipa.example",
              "replica1.ipa.example",
              "replica2.ipa.example"
            ]
          }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "dc=ipa,dc=example",
              "ruv": "3"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "o=ipaca",
              "ruv": "5"
            }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_dc=ipa,dc=example",
            "suffix": "dc=ipa,dc=example",
            "ruvs": [
              [
                "replica1.ipa.example",
                "3"
              ],
              [
                "ipa.ipa.example",
                "4"
              ],
              [
                "replica2.ipa.example",
                "7"
              ]
            ]
          }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_o=ipaca",
            "suffix": "o=ipaca",
            "ruvs": [
              [
                "replica1.ipa.example",
                "5"
              ],
              [
                "ipa.ipa.example",
                "6"
              ],
              [
                "replica2.ipa.example",
                "8"
              ]
            ]
          }
        }
    ],
    'replica2.ipa.example': [
        {
            "source": "ipahealthcheck.meta.core",
            "check": "MetaCheck",
            "result": "SUCCESS",
            "kw": {
              "fqdn": "replica2.ipa.example",
              "masters": [
                "ipa.ipa.example",
                "replica1.ipa.example",
                "replica2.ipa.example",
              ],
              "ipa_version": "4.8.4",
              "ipa_api_version": "2.235"
            }
        },
        {
          "source": "ipahealthcheck.ipa.meta",
          "check": "IPAMetaCheck",
          "result": "SUCCESS",
          "kw": {
            "masters": [
              "ipa.ipa.example",
              "replica1.ipa.example",
              "replica2.ipa.example"
            ]
          }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "dc=ipa,dc=example",
              "ruv": "7"
            }
        },
        {
            "source": "ipahealthcheck.ds.ruv",
            "check": "RUVCheck",
            "result": "SUCCESS",
            "kw": {
              "key": "o=ipaca",
              "ruv": "8"
            }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_dc=ipa,dc=example",
            "suffix": "dc=ipa,dc=example",
            "ruvs": [
              [
                "replica2.ipa.example",
                "7"
              ],
              [
                "ipa.ipa.example",
                "4"
              ],
              [
                "replica1.ipa.example",
                "3"
              ]
            ]
          }
        },
        {
          "source": "ipahealthcheck.ds.ruv",
          "check": "KnownRUVCheck",
          "result": "SUCCESS",
          "kw": {
            "key": "ruvs_o=ipaca",
            "suffix": "o=ipaca",
            "ruvs": [
              [
                "replica2.ipa.example",
                "8"
              ],
              [
                "ipa.ipa.example",
                "6"
              ],
              [
                "replica1.ipa.example",
                "5"
              ],
              [
                "replica1.ipa.example",
                "9"
              ]
            ]
          }
        }
    ]
}



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#
import re

from dns import (
    rdata,
    rdataclass,
    rdatatype,
    message,
    rrset,
    version,
)
from dns.resolver import Answer
from dns.exception import DNSException

from base import BaseTest
from util import capture_results, m_api
from unittest.mock import patch

from ipahealthcheck.core import config, constants
from ipahealthcheck.ipa.plugin import registry
from ipahealthcheck.ipa.idns import IPADNSSystemRecordsCheck

from ipapython.dnsutil import DNSName
from ipaserver.dns_data_management import (
    IPA_DEFAULT_MASTER_SRV_REC,
    IPA_DEFAULT_ADTRUST_SRV_REC
)

try:
    # pylint: disable=unused-import
    from ipaserver.dns_data_management import IPA_DEFAULT_MASTER_URI_REC  # noqa
except ImportError:
    has_uri_support = False
else:
    has_uri_support = True


try:
    # pylint: disable=unused-import
    from ipaserver.install.installutils import resolve_rrsets_nss  # noqa: F401
    # pylint: enable=unused-import
except ImportError:
    resolve_rrsets_import = 'ipaserver.dns_data_management.resolve_rrsets'
else:
    resolve_rrsets_import = 'ipaserver.install.installutils.resolve_rrsets_nss'

# Global for test_dnsrecords_missing_ca()
no_ca_count = 0

# Global for test_dnsrecords_extra_ca()
extra_ca_count = 0


def add_srv_records(qname, port_map, priority=0, weight=100):
    rdlist = []
    for _unused, port in port_map:
        answerlist = []
        for host in qname:
            hostname = DNSName(host)
            rd = rdata.from_text(
                rdataclass.IN, rdatatype.SRV,
                '{0} {1} {2} {3}'.format(
                    priority, weight, port, hostname.make_absolute()
                )
            )
            answerlist.append(rd)
        rdlist.append(answerlist)
    return rdlist


def resolve_rrsets(fqdn, rdtypes):
    """
    Return an A record for the hostname in an RRset type in a list.
    """
    rset = []
    for rdtype in rdtypes:
        rlist = rrset.from_text_list(fqdn, 86400, rdataclass.IN,
                                     rdtype, gen_addrs(rdtype, 1))
        rset.append(rlist)

    return rset


def query_srv(qname, ad_records=False):
    """
    Return a SRV for each service IPA cares about for all the hosts.

    This is pre-generated as a side-effect for each test.
    """
    rdlist = add_srv_records(qname, IPA_DEFAULT_MASTER_SRV_REC)
    if ad_records:
        rdlist.extend(add_srv_records(qname, IPA_DEFAULT_ADTRUST_SRV_REC))
    return rdlist


def query_uri(hosts):
    """
    Return a list containing two answers, one for each uri type
    """
    answers = []
    if version.MAJOR < 2 or (version.MAJOR == 2 and version.MINOR == 0):
        m = message.Message()
    elif version.MAJOR == 2 and version.MINOR > 0:
        m = message.QueryMessage()
        m = message.make_response(m)

    rdtype = rdatatype.URI
    for name in ('_kerberos.', '_kpasswd.'):
        qname = DNSName(name + m_api.env.domain)
        qname = qname.make_absolute()
        if version.MAJOR < 2:
            # pylint: disable=unexpected-keyword-arg
            answer = Answer(qname, rdataclass.IN, rdtype, m,
                            raise_on_no_answer=False)
            # pylint: enable=unexpected-keyword-arg
        else:
            if version.MAJOR == 2 and version.MINOR > 0:
                question = rrset.RRset(qname, rdataclass.IN, rdtype)
                m.question = [question]
            answer = Answer(qname, rdataclass.IN, rdtype, m)

        rl = []
        for host in hosts:
            rlist = rrset.from_text_list(
                qname, 86400, rdataclass.IN,
                rdatatype.URI, ['0 100 "krb5srv:m:tcp:%s."' % host,
                                '0 100 "krb5srv:m:udp:%s."' % host, ]
            )
            rl.extend(rlist)
        answer.rrset = rl
        answers.append(answer)
    return answers


def gen_addrs(rdtype=rdatatype.A, num=1):
    """Generate sequential IP addresses for the ipa-ca A record lookup"""
    ips = []
    if rdtype == rdatatype.A:
        ip_template = '192.168.0.%d'
    if rdtype == rdatatype.AAAA:
        ip_template = '2001:db8:1::%d'

    for i in range(num):
        ips.append(ip_template % (i + 1))

    return ips


def fake_query(qname, rdtype=rdatatype.A, rdclass=rdataclass.IN, count=1,
               fake_txt=False):
    """Fake a DNS query, returning count responses to the request

       Three kinds of lookups are faked:
       1. A query for A/AAAA records for a service will return the count
          as requested in the test. This simulates lookups for the
          ipa-ca A/AAAA record. To force a difference in responses one can
          vary the count.
       2. TXT queries will return the Kerberos realm

       fake_txt will set an invalid Kerberos realm entry to provoke a
       warning.
    """
    if version.MAJOR < 2 or (version.MAJOR == 2 and version.MINOR == 0):
        m = message.Message()
    elif version.MAJOR == 2 and version.MINOR > 0:
        m = message.QueryMessage()
        m = message.make_response(m)
    if rdtype in (rdatatype.A, rdatatype.AAAA):
        fqdn = DNSName(qname)
        fqdn = fqdn.make_absolute()

        if version.MAJOR < 2:
            # pylint: disable=unexpected-keyword-arg
            answers = Answer(fqdn, rdataclass.IN, rdtype, m,
                             raise_on_no_answer=False)
            # pylint: enable=unexpected-keyword-arg
        else:
            if version.MAJOR == 2 and version.MINOR > 0:
                question = rrset.RRset(fqdn, rdataclass.IN, rdtype)
                m.question = [question]
            answers = Answer(fqdn, rdataclass.IN, rdtype, m)

        rlist = rrset.from_text_list(fqdn, 86400, rdataclass.IN,
                                     rdtype, gen_addrs(rdtype, count))

        answers.rrset = rlist
    elif rdtype == rdatatype.TXT:
        if fake_txt:
            realm = 'FAKE_REALM'
        else:
            realm = m_api.env.realm
        qname = DNSName('_kerberos.' + m_api.env.domain)
        qname = qname.make_absolute()

        if version.MAJOR < 2:
            # pylint: disable=unexpected-keyword-arg
            answers = Answer(qname, rdataclass.IN, rdatatype.TXT, m,
                             raise_on_no_answer=False)
            # pylint: enable=unexpected-keyword-arg
        else:
            if version.MAJOR == 2 and version.MINOR > 0:
                question = rrset.RRset(qname, rdataclass.IN, rdtype)
                m.question = [question]
            answers = Answer(qname, rdataclass.IN, rdatatype.TXT, m)

        rlist = rrset.from_text_list(qname, 86400, rdataclass.IN,
                                     rdatatype.TXT, [realm])

        answers.rrset = rlist

    return answers


# Helpers to generate an appropriate number of A records for the
# ipa-ca and Kerberos realm responses. Optionally return a bogus
# TXT record.
def fake_query_one(qname, rdtype=rdatatype.A, rdclass=rdataclass.IN,
                   count=1):
    return fake_query(qname, rdtype, rdclass, count)


def fake_query_two(qname, rdtype=rdatatype.A, rdclass=rdataclass.IN,
                   count=2):
    return fake_query(qname, rdtype, rdclass, count)


def fake_query_three(qname, rdtype=rdatatype.A, rdclass=rdataclass.IN,
                     count=3):
    return fake_query(qname, rdtype, rdclass, count)


def fake_query_one_txt(qname, rdtype=rdatatype.A, rdclass=rdataclass.IN,
                       count=1):
    return fake_query(qname, rdtype, rdclass, count, fake_txt=True)


def fake_query_one_no_ca(qname, rdtype=rdatatype.A,
                         rdclass=rdataclass.IN, count=1):
    """This is the singular side for resolve(). Since I use the
       side_effect as a callable I need to maintain my own state
       so I can still take in arguments from the check.

       The idea here is to start raising exceptions when the
       CA servers are resolved in the system records such that
       a CA is droppped.
    """
    global no_ca_count
    no_ca_count += 1
    if no_ca_count == 3:
        # This drops the AAAA ipa-ca record
        raise DNSException()
    return fake_query(qname, rdtype, rdclass, count=1)


def fake_query_one_extra_ca(qname, rdtype=rdatatype.A,
                            rdclass=rdataclass.IN, count=1):
    """This is the singular side for resolve(). Since I use the
       side_effect as a callable I need to maintain my own state
       so I can still take in arguments from the check.

       This adds an extra record during the ipa-ca AAAA
       lookup.
    """
    global extra_ca_count
    extra_ca_count += 1
    if extra_ca_count == 3:
        return fake_query(qname, rdtype, rdclass, count=2)
    else:
        return fake_query(qname, rdtype, rdclass, count=1)


def get_results_by_severity(results, severity):
    """Return the results with a matching severity"""
    new_results = []
    for result in results:
        if result.result == severity:
            new_results.append(result)
    return new_results


class TestDNSSystemRecords(BaseTest):
    """Test that the SRV records checks are working properly

       The intention was to not override IPASystemRecords since
       this is the core mechanism that IPA uses to determine what
       recoreds should exist.

       Instead the DNS lookups are managed. This is done in two
       ways:

       1. The query_srv() override returns the set of configured
          servers for each type of SRV record.
       2. fake_query() overrides ipahealthcheck.ipa.idns.resolve to simulate
          A, AAAA and TXT record lookups.
    """
    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_single(self, mock_query, mock_query_uri,
                               mock_query_srv, mock_rrset):
        """Test single CA master, all SRV records"""
        mock_query.side_effect = fake_query_one
        mock_query_srv.side_effect = query_srv([m_api.env.host])
        mock_query_uri.side_effect = query_uri([m_api.env.host])
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA))
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
            ]
        }]
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 13
        else:
            expected = 9
        assert len(self.results) == expected

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.idns'
            assert result.check == 'IPADNSSystemRecordsCheck'

    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_two(self, mock_query, mock_query_uri,
                            mock_query_srv, mock_rrset):
        """Test two CA masters, all SRV records"""
        mock_query_srv.side_effect = query_srv([
            m_api.env.host,
            'replica.' + m_api.env.domain
        ])
        mock_query_uri.side_effect = query_uri([
            m_api.env.host,
            'replica.' + m_api.env.domain
        ])
        mock_query.side_effect = fake_query_two
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA)),
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
            ]
        }]

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 24
        else:
            expected = 16
        assert len(self.results) == expected

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.idns'
            assert result.check == 'IPADNSSystemRecordsCheck'

    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_three(self, mock_query, mock_query_uri,
                              mock_query_srv, mock_rrset):
        """Test three CA masters, all SRV records"""
        mock_query_srv.side_effect = query_srv([
            m_api.env.host,
            'replica.' + m_api.env.domain,
            'replica2.' + m_api.env.domain
        ])
        mock_query_uri.side_effect = query_uri([
            m_api.env.host,
            'replica.' + m_api.env.domain,
            'replica2.' + m_api.env.domain
        ])
        mock_query.side_effect = fake_query_three
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica2.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA)),
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica2.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
            ]
        }]

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 35
        else:
            expected = 23
        assert len(self.results) == expected

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.idns'
            assert result.check == 'IPADNSSystemRecordsCheck'

    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_three_mixed(self, mock_query, mock_query_uri,
                                    mock_query_srv, mock_rrset):
        """Test three masters, only one with a CA, all SRV records"""
        mock_query_srv.side_effect = query_srv([
            m_api.env.host,
            'replica.' + m_api.env.domain,
            'replica2.' + m_api.env.domain
        ])
        mock_query_uri.side_effect = query_uri([
            m_api.env.host,
            'replica.' + m_api.env.domain,
            'replica2.' + m_api.env.domain
        ])
        mock_query.side_effect = fake_query_one
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica2.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA))
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica2.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'IPA master'
                    ],
                },
            ]
        }]

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 35
        else:
            expected = 23
        assert len(self.results) == expected

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.idns'

    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_missing_server(self, mock_query, mock_query_uri,
                                       mock_query_srv, mock_rrset):
        """Drop one of the masters from query_srv

           This will simulate missing SRV records and cause a number of
           warnings to be thrown.
        """
        mock_query_srv.side_effect = query_srv([
            m_api.env.host,
            'replica.' + m_api.env.domain
            # replica2 is missing
        ])
        mock_query_uri.side_effect = query_uri([
            m_api.env.host,
            'replica.' + m_api.env.domain,
            'replica2.' + m_api.env.domain
        ])
        mock_query.side_effect = fake_query_three
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica2.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA)),
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica2.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
            ]
        }]

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 35
        else:
            expected = 23
        assert len(self.results) == expected

        ok = get_results_by_severity(self.results.results, constants.SUCCESS)
        warn = get_results_by_severity(self.results.results, constants.WARNING)
        if has_uri_support:
            assert len(ok) == 28
            assert len(warn) == 7
        else:
            assert len(ok) == 16
            assert len(warn) == 7

        for result in warn:
            assert result.kw.get('msg') == 'Expected SRV record missing'

    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_missing_ipa_ca(self, mock_query, mock_query_uri,
                                       mock_query_srv, mock_rrset):
        """Drop one of the masters from query_srv

           This will simulate missing SRV records and cause a number of
           warnings to be thrown.
        """
        mock_query_srv.side_effect = query_srv([
            m_api.env.host,
            'replica.' + m_api.env.domain,
            'replica2.' + m_api.env.domain
        ])
        mock_query_uri.side_effect = query_uri([
            m_api.env.host,
            'replica.' + m_api.env.domain,
            'replica2.' + m_api.env.domain
        ])
        mock_query.side_effect = fake_query_two
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica2.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA))
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica2.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
            ]
        }]

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 35
        else:
            expected = 23
        assert len(self.results) == expected

        ok = get_results_by_severity(self.results.results, constants.SUCCESS)
        warn = get_results_by_severity(self.results.results, constants.WARNING)
        if has_uri_support:
            assert len(ok) == 35
            assert len(warn) == 0
        else:
            assert len(ok) == 23
            assert len(warn) == 0

        for result in warn:
            assert re.match(
                r'^Got {count} ipa-ca (A|AAAA) records, expected {expected}$',
                result.kw.get('msg')
            )
            assert result.kw.get('count') == 2
            assert result.kw.get('expected') == 3

    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_extra_srv(self, mock_query, mock_query_uri,
                                  mock_query_srv, mock_rrset):
        """An extra SRV record set exists, report it.

           Add an extra master to the query_srv() which will generate
           a full extra set of SRV records for the master.
        """
        mock_query_srv.side_effect = query_srv([
            m_api.env.host,
            'replica.' + m_api.env.domain,
            'replica2.' + m_api.env.domain,
            'replica3.' + m_api.env.domain
        ])
        mock_query_uri.side_effect = query_uri([
            m_api.env.host,
            'replica.' + m_api.env.domain,
            'replica2.' + m_api.env.domain,
        ])
        mock_query.side_effect = fake_query_three
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica2.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA)),
            resolve_rrsets('replica3.' + m_api.env.domain,
                           (rdatatype.A, rdatatype.AAAA)),
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
                {
                    'cn': ['replica2.' + m_api.env.domain],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
            ]
        }]

        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 42
        else:
            expected = 30
        assert len(self.results) == expected

        ok = get_results_by_severity(self.results.results, constants.SUCCESS)
        warn = get_results_by_severity(self.results.results, constants.WARNING)
        if has_uri_support:
            assert len(ok) == 35
            assert len(warn) == 7
        else:
            assert len(ok) == 23
            assert len(warn) == 7

        for result in warn:
            assert result.kw.get('msg') == \
                'Unexpected SRV entry in DNS'

    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_bad_realm(self, mock_query, mock_query_uri,
                                  mock_query_srv, mock_rrset):
        """Unexpected Kerberos TXT record"""
        mock_query.side_effect = fake_query_one_txt
        mock_query_srv.side_effect = query_srv([m_api.env.host])
        mock_query_uri.side_effect = query_uri([m_api.env.host])
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA))
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
            ]
        }]
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 13
        else:
            expected = 9
        assert len(self.results) == expected

        ok = get_results_by_severity(self.results.results, constants.SUCCESS)
        warn = get_results_by_severity(self.results.results, constants.WARNING)
        if has_uri_support:
            assert len(ok) == 12
            assert len(warn) == 1
        else:
            assert len(ok) == 8
            assert len(warn) == 1

        result = warn[0]
        assert result.kw.get('msg') == 'expected realm missing'
        assert result.kw.get('key') == '\"FAKE_REALM\"'

    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_one_with_ad(self, mock_query, mock_query_uri,
                                    mock_query_srv, mock_rrset):
        mock_query.side_effect = fake_query_one
        mock_query_srv.side_effect = query_srv([m_api.env.host], True)
        mock_query_uri.side_effect = query_uri([m_api.env.host])
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA))
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master',
                        'AD trust controller'
                    ],
                },
            ]
        }]
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 19
        else:
            expected = 15
        assert len(self.results) == expected

        for result in self.results.results:
            assert result.result == constants.SUCCESS
            assert result.source == 'ipahealthcheck.ipa.idns'
            assert result.check == 'IPADNSSystemRecordsCheck'

    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_missing_ca(self, mock_query, mock_query_uri,
                                   mock_query_srv, mock_rrset):
        """Missing AAAA record in ipa-ca"""
        mock_query.side_effect = fake_query_one_no_ca
        mock_query_srv.side_effect = query_srv([m_api.env.host])
        mock_query_uri.side_effect = query_uri([m_api.env.host])
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA,))
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
            ]
        }]
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 13
        else:
            expected = 8
        assert len(self.results) == expected

        ok = get_results_by_severity(self.results.results, constants.SUCCESS)
        warn = get_results_by_severity(self.results.results, constants.WARNING)
        if has_uri_support:
            assert len(ok) == 12
            assert len(warn) == 1
        else:
            assert len(ok) == 8
            assert len(warn) == 1

        result = warn[0]
        assert result.kw.get('msg') == (
            'expected ipa-ca to contain {ipaddr} for {server}'
        )
        assert result.kw.get('ipaddr') == '2001:db8:1::1'

    @patch(resolve_rrsets_import)
    @patch('ipapython.dnsutil.query_srv')
    @patch('ipahealthcheck.ipa.idns.query_uri')
    @patch('ipahealthcheck.ipa.idns.resolve')
    def test_dnsrecords_extra_ca(self, mock_query, mock_query_uri,
                                 mock_query_srv, mock_rrset):
        """Extra AAAA record in ipa-ca"""
        mock_query.side_effect = fake_query_one_extra_ca
        mock_query_srv.side_effect = query_srv([m_api.env.host])
        mock_query_uri.side_effect = query_uri([m_api.env.host])
        mock_rrset.side_effect = [
            resolve_rrsets(m_api.env.host, (rdatatype.A, rdatatype.AAAA,))
        ]

        m_api.Command.server_find.side_effect = [{
            'result': [
                {
                    'cn': [m_api.env.host],
                    'enabled_role_servrole': [
                        'CA server',
                        'IPA master'
                    ],
                },
            ]
        }]
        framework = object()
        registry.initialize(framework, config.Config)
        f = IPADNSSystemRecordsCheck(registry)

        self.results = capture_results(f)

        if has_uri_support:
            expected = 13
        else:
            expected = 8
        assert len(self.results) == expected

        ok = get_results_by_severity(self.results.results, constants.SUCCESS)
        warn = get_results_by_severity(self.results.results, constants.WARNING)
        if has_uri_support:
            assert len(ok) == 12
            assert len(warn) == 1
        else:
            assert len(ok) == 8
            assert len(warn) == 1

        result = warn[0]
        assert result.kw.get('msg') == (
            'Unexpected ipa-ca address {ipaddr}'
        )
        assert result.kw.get('ipaddr') == '2001:db8:1::2'



#
# Copyright (C) 2020 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.ds.plugin import DSPlugin, registry
from lib389.monitor import MonitorDiskSpace


@registry
class DiskSpaceCheck(DSPlugin):
    """
    Check the all the disks that the DS uses
    """
    check_class = MonitorDiskSpace



#
# Copyright (C) 2020 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.ds.plugin import DSPlugin, registry
from lib389.nss_ssl import NssSsl


@registry
class NssCheck(DSPlugin):
    """
    Check the NSS database certificates for expiring issues
    """
    check_class = NssSsl



#
# Copyright (C) 2020 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.ds.plugin import DSPlugin, registry
from lib389.backend import Backends


@registry
class BackendsCheck(DSPlugin):
    """
    Check all the backends for misconfigurations
    """
    check_class = Backends
    many = True



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import logging

from ipahealthcheck.ipa.plugin import IPAPlugin, registry
from ipahealthcheck.core.plugin import Result
from ipahealthcheck.core.plugin import duration
from ipahealthcheck.core import constants

from ipalib import api

logger = logging.getLogger()


@registry
class IPACRLManagerCheck(IPAPlugin):
    """
    Determine if this master is the CRL manager

    This check in itself will always return SUCCESS and is only
    useful in the context of the ohter masters. Some external
    service is expected to aggregate this.
    """
    @duration
    def check(self):
        if not self.ca.is_configured():
            return
        try:
            enabled = self.ca.is_crlgen_enabled()
        except AttributeError:
            yield Result(self, constants.SUCCESS,
                         key='crl_manager',
                         crlgen_enabled=None,
                         msg='Not available in this version of IPA')
        else:
            yield Result(self, constants.SUCCESS,
                         key='crl_manager',
                         crlgen_enabled=enabled)


@registry
class IPARenewalMasterCheck(IPAPlugin):
    """
    Determine if this master is the CA renewal master.

    This check in itself will always return SUCCESS and is only
    useful in the context of the ohter masters. Some external
    service is expected to aggregate this.
    """
    requires = ('dirsrv',)

    @duration
    def check(self):
        try:
            result = api.Command.config_show()
        except Exception as e:
            yield Result(self, constants.ERROR,
                         key='renewal_master',
                         msg='Request for configuration failed, %s' % e)
        else:
            server = result['result'].get('ca_renewal_master_server', None)
            yield Result(self, constants.SUCCESS,
                         key='renewal_master',
                         master=server == api.env.host)




# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import gssapi
import logging
import os
import tempfile

from ipahealthcheck.ipa.plugin import IPAPlugin, registry
from ipahealthcheck.core.plugin import Result, duration
from ipahealthcheck.core import constants

from ipalib import api
from ipalib.install.kinit import kinit_keytab
from ipaplatform.paths import paths
from ipapython import ipautil


logger = logging.getLogger()


@registry
class IPAHostKeytab(IPAPlugin):
    """Ensure the host keytab can get a TGT"""
    requires = ('krb5kdc', 'dirsrv')

    @duration
    def check(self):
        ccache_dir = tempfile.mkdtemp()
        ccache_name = os.path.join(ccache_dir, 'ccache')

        try:
            try:
                host_princ = str('host/%s@%s' % (api.env.host, api.env.realm))
                kinit_keytab(host_princ, paths.KRB5_KEYTAB, ccache_name)
            except gssapi.exceptions.GSSError as e:
                yield Result(self, constants.ERROR,
                             msg='Failed to obtain host TGT: %s' % e)
        finally:
            ipautil.remove_file(ccache_name)
            os.rmdir(ccache_dir)



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

import glob
import logging
import os

from ipahealthcheck.core.files import FileCheck
from ipahealthcheck.ipa.plugin import IPAPlugin, registry

from ipalib import api, errors

from ipaplatform.paths import paths
from ipaplatform.constants import constants
from ipapython.certdb import NSS_SQL_FILES
from ipapython.dn import DN

from ipaserver.install import dsinstance
from ipaserver.install import krbinstance


logger = logging.getLogger()


@registry
class IPAFileNSSDBCheck(IPAPlugin, FileCheck):

    def collect_files(self, basedir, filelist, owner, group, perms):
        for file in filelist:
            self.files.append((os.path.join(basedir, file), owner, group,
                              perms))

    def check(self):
        self.files = []

        self.collect_files(dsinstance.config_dirname(self.serverid),
                           NSS_SQL_FILES, constants.DS_USER, 'root', '0640')

        # There always has to be a special one. pkcs11.txt has a different
        # group so pop off the auto-generated one and add a replacement.
        old = (os.path.join(dsinstance.config_dirname(self.serverid),
               'pkcs11.txt'), constants.DS_USER, 'root', '0640')
        self.files.remove(old)
        new = (os.path.join(dsinstance.config_dirname(self.serverid),
               'pkcs11.txt'), constants.DS_USER, constants.DS_GROUP, '0640')
        self.files.append(new)

        if self.ca.is_configured():
            self.collect_files(paths.PKI_TOMCAT_ALIAS_DIR, NSS_SQL_FILES,
                               constants.PKI_USER, constants.PKI_GROUP, '0600')

        return FileCheck.check(self)


@registry
class IPAFileCheck(IPAPlugin, FileCheck):
    def dns_container_exists(self):
        try:
            self.conn.get_entry(DN(api.env.container_dns,
                                api.env.basedn), [])
        except errors.NotFound:
            return False
        except AttributeError:
            logger.debug("LDAP is down, can't tell whether DNS is available."
                         " Skipping those file checks.")
            return False
        return True

    def check(self):
        self.files = []

        if self.ca.is_configured():
            self.files.append(
                (paths.RA_AGENT_PEM, 'root', constants.IPAAPI_GROUP, '0440')
            )
            self.files.append(
                (paths.RA_AGENT_KEY, 'root', constants.IPAAPI_GROUP, '0440')
            )

        if krbinstance.is_pkinit_enabled():
            self.files.append((paths.KDC_CERT, 'root', 'root', '0644'))
            self.files.append((paths.KDC_KEY, 'root', 'root', '0600'))

        if self.dns_container_exists():
            self.files.append((paths.NAMED_KEYTAB,
                              constants.NAMED_USER,
                              constants.NAMED_GROUP, '0400'))
            if os.path.exists(paths.IPA_DNSKEYSYNCD_KEYTAB):
                self.files.append((paths.IPA_DNSKEYSYNCD_KEYTAB,
                                  'root', constants.ODS_GROUP, '0440'))

        self.files.append((paths.GSSAPI_SESSION_KEY,
                          'root', 'root', '0600'))
        self.files.append((paths.DS_KEYTAB,
                           constants.DS_USER, constants.DS_GROUP, '0600'))
        self.files.append((paths.IPA_CA_CRT, 'root', 'root', '0644'))
        self.files.append((paths.IPA_CUSTODIA_KEYS, 'root', 'root', '0600'))

        self.files.append((paths.RESOLV_CONF, ('root', 'systemd-resolve'),
                          ('root', 'systemd-resolve'), '0644'))
        self.files.append((paths.HOSTS, 'root', 'root', '0644'))

        # IPA log files that may vary by installation. Only verify
        # those that exist
        for filename in (
            paths.IPABACKUP_LOG,
            paths.IPARESTORE_LOG,
            paths.IPACLIENT_INSTALL_LOG,
            paths.IPACLIENT_UNINSTALL_LOG,
            paths.IPAREPLICA_CA_INSTALL_LOG,
            paths.IPAREPLICA_CONNCHECK_LOG,
            paths.IPAREPLICA_INSTALL_LOG,
            paths.IPASERVER_INSTALL_LOG,
            paths.IPASERVER_KRA_INSTALL_LOG,
            paths.IPASERVER_UNINSTALL_LOG,
            paths.IPAUPGRADE_LOG,
            paths.IPATRUSTENABLEAGENT_LOG,
        ):
            if os.path.exists(filename):
                self.files.append((filename, 'root', 'root', '0600'))

        self.files.append((paths.IPA_CUSTODIA_AUDIT_LOG,
                          'root', 'root', '0644'))

        self.files.append((paths.KADMIND_LOG, 'root', 'root',
                          ('0600', '0640')))
        self.files.append((paths.KRB5KDC_LOG, 'root', 'root', '0640'))

        inst = api.env.realm.replace('.', '-')
        self.files.append((paths.SLAPD_INSTANCE_ACCESS_LOG_TEMPLATE % inst,
                           constants.DS_USER, constants.DS_GROUP, '0600'))
        self.files.append((paths.SLAPD_INSTANCE_ERROR_LOG_TEMPLATE % inst,
                           constants.DS_USER, constants.DS_GROUP, '0600'))

        self.files.append((paths.VAR_LOG_HTTPD_ERROR, 'root', 'root', '0644'))

        for globpath in glob.glob("%s/debug*.log" % paths.TOMCAT_CA_DIR):
            self.files.append(
                (globpath, constants.PKI_USER, constants.PKI_GROUP, "0644")
            )

        for globpath in glob.glob(
            "%s/ca_audit*" % paths.TOMCAT_SIGNEDAUDIT_DIR
        ):
            self.files.append(
                (globpath, constants.PKI_USER, constants.PKI_GROUP, '0640')
            )

        for filename in ('selftests.log', 'system', 'transactions'):
            self.files.append((
                os.path.join(paths.TOMCAT_CA_DIR, filename),
                constants.PKI_USER, constants.PKI_GROUP, '0640'
            ))

        for globpath in glob.glob("%s/debug*.log" % paths.TOMCAT_KRA_DIR):
            self.files.append(
                (globpath, constants.PKI_USER, constants.PKI_GROUP, "0644")
            )

        for globpath in glob.glob(
            "%s/ca_audit*" % paths.TOMCAT_KRA_SIGNEDAUDIT_DIR
        ):
            self.files.append(
                (globpath, constants.PKI_USER, constants.PKI_GROUP, '0640')
            )

        for filename in ('selftests.log', 'system', 'transactions'):
            self.files.append((
                os.path.join(paths.TOMCAT_KRA_DIR, filename),
                constants.PKI_USER, constants.PKI_GROUP, '0640'
            ))

        self.files.append((paths.IPA_CCACHES,
                           constants.IPAAPI_USER, constants.IPAAPI_GROUP,
                           '6770'))
        self.files.append((paths.IPA_RENEWAL_LOCK, 'root', 'root', '0600'))
        self.files.append((paths.SVC_LIST_FILE, 'root', 'root', '0644'))

        return FileCheck.check(self)


@registry
class TomcatFileCheck(IPAPlugin, FileCheck):
    def check(self):
        if not self.ca.is_configured():
            logger.debug('CA is not configured, skipping')
            self.files = []
        else:
            self.files = [
                (paths.PKI_TOMCAT_PASSWORD_CONF,
                 constants.PKI_USER, constants.PKI_GROUP, '0660'),
                (paths.CA_CS_CFG_PATH,
                 constants.PKI_USER, constants.PKI_GROUP, '0660'),
                (os.path.join(paths.PKI_TOMCAT, 'server.xml'),
                 constants.PKI_USER, constants.PKI_GROUP, '0660'),
            ]

        return FileCheck.check(self)



#
# Copyright (C) 2021 FreeIPA Contributors see COPYING for license
#

import grp
import logging

from ipaplatform.constants import constants as platform_constants

from ipahealthcheck.ipa.plugin import IPAPlugin, registry
from ipahealthcheck.core.plugin import Result
from ipahealthcheck.core.plugin import duration
from ipahealthcheck.core import constants

logger = logging.getLogger()

# A tuple of groups and a tuple of expected members
#
# For example the apache user needs to be in the ipaapi group so
# the tuple would look like: 'ipaapi', ('apache',).
#
# The second value is a tuple so that we can more easily extend if
# multiple users need to be a member of a group.
#
# (group_name, (members,))
GROUP_MEMBERS = (
    (platform_constants.IPAAPI_GROUP, (platform_constants.HTTPD_USER,)),
)


@registry
class IPAGroupMemberCheck(IPAPlugin):
    """
    Ensure that nss/POSIX group membership is as expected.

    This can be critical for security and/or proper access control and
    is primarily being checked for privilege separation. The ipaapi
    user needs to be able to read ccaches created by Apache.
    """
    @duration
    def check(self):
        for (group, members) in GROUP_MEMBERS:
            try:
                grp_group = grp.getgrnam(group)
            except KeyError:
                yield Result(self, constants.ERROR,
                             key=group,
                             msg='group {key} does not exist')
                continue
            for member in members:
                if member not in grp_group.gr_mem:
                    yield Result(self, constants.ERROR,
                                 key=group,
                                 member=member,
                                 msg='{member} is not a member of group {key}')
                else:
                    yield Result(self, constants.SUCCESS,
                                 key=group,
                                 member=member)



#
# Copyright (C) 2019 FreeIPA Contributors see COPYING for license
#

from ipahealthcheck.core.plugin import Plugin, Registry
from ipaserver.install import cainstance
from ipaserver.install import installutils
from ipalib import api, errors


class DogtagPlugin(Plugin):
    def __init__(self, reg):
        super().__init__(reg)
        self.ca = cainstance.CAInstance(api.env.realm, host_name=api.env.host)


class DogtagRegistry(Registry):
    def initialize(self, framework, config, options=None):
        super().initialize(framework, config)
        installutils.check_server_configuration()
        if not api.isdone('bootstrap'):
            api.bootstrap(in_server=True,
                          context='ipahealthcheck',
                          log=None)
        if not api.isdone('finalize'):
            api.finalize()

        if not api.Backend.ldap2.isconnected():
            try:
                api.Backend.ldap2.connect()
            except errors.CCacheError:
                pass
            except errors.NetworkError:
                pass


registry = DogtagRegistry()
